# Copyright 2010-2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from nose.tools import istest
from reahl.tofu import Fixture, test, vassert, expected, NoException, scenario
from reahl.stubble import EmptyStub

from reahl.web.fw import UserInterface
from reahl.web.fw import Region
from reahl.web.ui import HTML5Page, TwoColumnPage, P
from reahl.webdev.tools import Browser
from reahl.web_dev.fixtures import WebFixture, ReahlWSGIApplicationStub
from reahl.component.exceptions import ProgrammerError, IncorrectArgumentError, IsSubclass, IsInstance

@istest
class AppBasicsTests(object):
    class BasicScenarios(WebFixture):
        @scenario
        def view_with_page(self):
            class SimplePage(HTML5Page):
                def __init__(self, view):
                    super(SimplePage, self).__init__(view)
                    self.body.add_child(P(view, text=u'Hello world!'))

            class MainUI(UserInterface):
                def assemble(self):
                    self.define_view(u'/', title=u'Hello', page=SimplePage.factory())

            self.MainUI = MainUI
            self.expected_content_length = 685
            self.content_includes_p = True

        @scenario
        def view_with_set_page(self):
            class SimplePage(HTML5Page):
                def __init__(self, view):
                    super(SimplePage, self).__init__(view)
                    self.body.add_child(P(view, text=u'Hello world!'))

            class MainUI(UserInterface):
                def assemble(self):
                    home = self.define_view(u'/', title=u'Hello')
                    home.set_page(SimplePage.factory())

            self.MainUI = MainUI
            self.expected_content_length = 685
            self.content_includes_p = True

        @scenario
        def ui_with_page(self):
            class MainUI(UserInterface):
                def assemble(self):
                    self.define_page(TwoColumnPage)
                    self.define_view(u'/', title=u'Hello')

            self.MainUI = MainUI
            self.expected_content_length = 893
            self.content_includes_p = False

        @scenario
        def backwards_compatibility(self):
            class MainUI(Region):
                def assemble(self):
                    self.define_main_window(TwoColumnPage)
                    self.define_view(u'/', title=u'Hello')

            self.MainUI = MainUI
            self.expected_content_length = 893
            self.content_includes_p = False

    @test(BasicScenarios)
    def basic_assembly(self, fixture):
        """An application is built by extending UserInterface, and defining this UserInterface in an .assemble() method.

        To define the UserInterface, several Views are defined. Views are mapped to URLs. When a user GETs
        the URL of a View, a page is rendered back to the user. How that page is created
        can happen in different ways, as illustrated by each scenario of this test.
        """
        wsgi_app = fixture.new_wsgi_app(site_root=fixture.MainUI)
        browser = Browser(wsgi_app)

        # GETting the URL results in the HTML for that View
        browser.open('/')
        vassert( browser.title == u'Hello' )

        if fixture.content_includes_p:
            [message] = browser.lxml_html.xpath('//p')
            vassert( message.text == u'Hello world!' )

        # The headers are set correctly
        response = browser.last_response
        vassert( response.content_length == fixture.expected_content_length )
        vassert( response.content_type == 'text/html' )
        vassert( response.charset == 'utf-8' )

        # Invalid URLs do not exist
        browser.open('/nonexistantview/', status=404)



    @test(WebFixture)
    def basic_error1(self, fixture):
        """Sending the the wrong kind of thing as widget_class to define_page is reported to the programmer."""
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(EmptyStub)
                self.define_view(u'/', title=u'Hello')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        with expected(IsSubclass):
            browser.open('/')

    @test(WebFixture)
    def basic_error2(self, fixture):
        """Sending the the wrong arguments for the specified class to define_page is reported to the programmer."""
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage, 1, 2)
                self.define_view(u'/', title=u'Hello')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        def check_exc(ex):
            msg = str(ex)
            vassert( msg.startswith('define_page was called with arguments that do not match those expected by') )
        with expected(IncorrectArgumentError, test=check_exc):
            browser.open('/')

    @test(WebFixture)
    def basic_error3(self, fixture):
        """Forgetting to define either a page of a page for a View is reported to the programmer."""
        class MainUI(UserInterface):
            def assemble(self):
                self.define_view(u'/', title=u'Hello')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        def check_exc(ex):
            msg = str(ex)
            vassert( msg == 'there is no page defined for /' )
        with expected(ProgrammerError, test=check_exc):
            browser.open('/')

    class SlotScenarios(WebFixture):
        @scenario
        def page_on_ui(self):
            class MainUI(UserInterface):
                def assemble(self):
                    self.define_page(TwoColumnPage)
                    home = self.define_view(u'/', title=u'Hello')
                    home.set_slot(u'main', P.factory(text=u'Hello world'))
                    home.set_slot(u'footer', P.factory(text=u'I am the footer'))
            self.MainUI = MainUI

        @scenario
        def page_on_view(self):
            class MainUI(UserInterface):
                def assemble(self):
                    home = self.define_view(u'/', title=u'Hello')
                    home.set_page(TwoColumnPage.factory())
                    home.set_slot(u'main', P.factory(text=u'Hello world'))
                    home.set_slot(u'footer', P.factory(text=u'I am the footer'))
            self.MainUI = MainUI

            
    @test(SlotScenarios)
    def slots(self, fixture):
        """A View modifies the page by populating named Slots in the page with Widgets."""
        wsgi_app = fixture.new_wsgi_app(site_root=fixture.MainUI)
        browser = Browser(wsgi_app)
        
        browser.open('/')
        vassert( browser.title == u'Hello' )
        [main_p, footer_p] = browser.lxml_html.xpath('//p')
        vassert( main_p.text == u'Hello world' )
        vassert( footer_p.text == u'I am the footer' )

    @test(WebFixture)
    def slot_error(self, fixture):
        """Supplying contents for a slot that does not exist results in s sensible error."""
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/', title=u'Hello')
                home.set_slot(u'main', P.factory(text=u'Hello world'))
                home.set_slot(u'nonexistantslotname', P.factory(text=u'I am breaking'))

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        def check_exc(ex):
            vassert( str(ex).startswith(u'An attempt was made to plug Widgets into the following slots that do not exist') )

        with expected(ProgrammerError, test=check_exc):
            browser.open('/')


    @test(WebFixture)
    def slot_defaults(self, fixture):
        """If a View does not specify contents for a Slot, the Slot will be populated by the window's default
           widget for that slot if specified, else it will be left empty.
        """
        class MainUI(UserInterface):
            def assemble(self):
                main = self.define_page(TwoColumnPage)
                main.add_default_slot(u'main', P.factory(text=u'defaulted slot contents'))
                self.define_view(u'/', title=u'Hello')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)
        
        browser.open('/')

        # The default widget for the main slot is used
        [p] = browser.lxml_html.xpath('//p')
        vassert( p.text == 'defaulted slot contents' )

        # The header slot has no default, and is thus left empty
        header_contents = browser.lxml_html.xpath('//header/*')
        vassert( not header_contents )



# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from nose.tools import istest
from reahl.tofu import Fixture, test, scenario
from reahl.tofu import vassert, expected

from reahl.web.fw import ReahlWSGIApplication, UserInterface
from reahl.web.ui import TwoColumnPage, P, A
from reahl.webdev.tools import Browser, WidgetTester
from reahl.web_dev.fixtures import WebFixture
from reahl.component.exceptions import ProgrammerError

class UserInterfaceErrorScenarios(WebFixture):
    def new_wsgi_app(self):
        fixture = self
        class SimpleUserInterface(UserInterface):
            def assemble(self):
                root = self.define_view(u'/', title=u'View')
                root.set_slot(u'name', P.factory())

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  SimpleUserInterface,  {}, name=u'test_ui')

        return super(UserInterfaceErrorScenarios, self).new_wsgi_app(site_root=MainUI)

    @scenario
    def plug_in_to_nonexistant_name(self):
        self.slot_map = {u'name': u'nonexistent'}

    @scenario
    def view_name_not_mapped(self):
        self.slot_map = {}

        
@istest
class UserInterfaceErrorTests(object):
    @test(UserInterfaceErrorScenarios)
    def ui_slots_map_error(self, fixture):
        browser = Browser(fixture.wsgi_app)

        with expected(ProgrammerError):
            browser.open('/a_ui/')





# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import StringIO
import datetime
import os.path

import pkg_resources

from nose.tools import istest
from reahl.tofu import Fixture, test, scenario
from reahl.tofu import vassert, expected, temp_dir, temp_file_with
from reahl.stubble import easter_egg, stubclass

from reahl.web.fw import FileOnDisk, FileFromBlob, PackagedFile, ConcatenatedFile, FileDownload, ReahlWSGIApplication, UserInterface
from reahl.web.ui import TwoColumnPage, P
from reahl.webdev.tools import Browser
from reahl.web_dev.fixtures import WebFixture

@istest
class StaticFileTests(object):
    @test(WebFixture)
    def files_from_disk(self, fixture):
        """A directory in the web.static_root configuration setting, can be mounted on a URL
           named after it on the WebApplication.
        """
        static_root = temp_dir()
        files_dir = static_root.sub_dir(u'staticfiles')
        sub_dir = files_dir.sub_dir(u'subdir')
        one_file = files_dir.file_with(u'one_file.xml', u'one')
        nested_file = sub_dir.file_with(u'nested_file', u'other')

        # How the config is set
        fixture.config.web.static_root = static_root.name
        
        # How the subdirectory is mounted
        class MainUI(UserInterface):
            def assemble(self):
                self.define_static_directory(u'/staticfiles')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # How the first file would be accessed
        browser.open('/staticfiles/one_file.xml')
        vassert( browser.raw_html == u'one' )
        
        # The meta-info of the file
        response = browser.last_response
        vassert( response.content_type == u'application/xml' )
        vassert( response.content_encoding is None )
        vassert( response.content_length == 3 )
        expected_mtime = datetime.datetime.fromtimestamp(int(os.path.getmtime(one_file.name)))
        vassert( response.last_modified.replace(tzinfo=None) == expected_mtime )
        expected_tag = u'%s-%s-%s' % (os.path.getmtime(one_file.name), 3, abs(hash(one_file.name)))
        vassert( response.etag == expected_tag )

        # How the file in the subdirectory would be accessed
        browser.open('/staticfiles/subdir/nested_file')
        vassert( browser.raw_html == u'other' )
        
        # When a file does not exist
        browser.open('/staticfiles/one_that_does_not_exist', status=404)

    @test(WebFixture)
    def files_from_list(self, fixture):
        """An explicit list of files can also be added on an URL as if they were in a single
           directory.
        """
        files_dir = temp_dir()
        one_file = files_dir.file_with(u'any_name_will_do_here', u'one')
        
        class MainUI(UserInterface):
            def assemble(self):
                list_of_files = [FileOnDisk(one_file.name, u'one_file')]
                self.define_static_files(u'/morestaticfiles', list_of_files)

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # How the first file would be accessed
        browser.open('/morestaticfiles/one_file')
        vassert( browser.raw_html == u'one' )
        
        # The meta-info of the file
        response = browser.last_response
        vassert( response.content_type == u'application/octet-stream' )
        vassert( response.content_encoding is None )
        vassert( response.content_length == 3 )
        expected_mtime = datetime.datetime.fromtimestamp(int(os.path.getmtime(one_file.name)))
        vassert( response.last_modified.replace(tzinfo=None) == expected_mtime )
        expected_tag = u'%s-%s-%s' % (os.path.getmtime(one_file.name), 3, abs(hash(one_file.name)))
        vassert( response.etag == expected_tag )

        # When a file does not exist
        browser.open('/morestaticfiles/one_that_does_not_exist', status=404)

    @test(WebFixture)
    def files_from_database(self, fixture):
        """Files can also be created on the fly such as from data in a database."""

        class MainUI(UserInterface):
            def assemble(self):
                content_type = u'text/html'
                encoding = u'utf-8'
                size = 10
                mtime = 123
                meta_info = content_type, encoding, size, mtime
                data_blob = StringIO.StringIO(u'x'*size)

                list_of_files = [FileFromBlob(u'database_file', data_blob, *meta_info)]
                self.define_static_files(u'/files', list_of_files)

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # How the file would be accessed
        browser.open('/files/database_file')
        vassert( browser.raw_html == u'xxxxxxxxxx' )
        response = browser.last_response

        # The meta-info of the file
        vassert( response.content_type == u'text/html' )
        vassert( response.content_encoding == u'utf-8' )
        vassert( response.content_length == 10 )
        vassert( response.last_modified.replace(tzinfo=None) == datetime.datetime.fromtimestamp(123) )
        expected_etag = '123-10-%s' % abs(hash(u'database_file'))
        vassert( response.etag == expected_etag )

    @test(WebFixture)
    def packaged_files(self, fixture):
        """Files can also be served straight from a python egg."""

        # Create an egg with package packaged_files, containing the file packaged_file
        egg_dir = temp_dir()
        package_dir = egg_dir.sub_dir(u'packaged_files')
        init_file = package_dir.file_with(u'__init__.py', u'')
        afile = package_dir.file_with(u'packaged_file', u'contents')

        pkg_resources.working_set.add(easter_egg)
        easter_egg.set_module_path(egg_dir.name)
        
        class MainUI(UserInterface):
            def assemble(self):
                list_of_files = [PackagedFile(easter_egg.as_requirement_string(), u'packaged_files', u'packaged_file')]
                self.define_static_files(u'/files', list_of_files)

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # How the file would be accessed
        browser.open('/files/packaged_file')
        vassert( browser.raw_html == u'contents' )

    class ConcatenateScenarios(WebFixture):
        @scenario
        def normal_files(self):
            self.file1_contents = u'contents1'
            self.file2_contents = u'contents2'
            self.filename = u'concatenated'
            self.expected_result = u'contents1contents2'

        @scenario
        def javascript_files(self):
            self.file1_contents = u'acall1(); //some comment'
            self.file2_contents = u'acall2(); //some comment'
            self.filename = u'concatenated.js'
            self.expected_result = u'acall1();acall2();'

        @scenario
        def css_files(self):
            self.file1_contents = u'.cool {}/* a comment */ '
            self.file2_contents = u'a, p { } '
            self.filename = u'concatenated.css'
            self.expected_result = u'.cool{}a,p{}'
            
    @test(ConcatenateScenarios)
    def concatenated_files(self, fixture):
        """Files can also be formed by concatenating other files.  Files ending in .css or .js are appropriately 
           minified in the process."""

        # Make an egg with a package called packaged_files, and two files in there.
        egg_dir = temp_dir()
        package_dir = egg_dir.sub_dir(u'packaged_files')
        init_file = package_dir.file_with(u'__init__.py', u'')
        afile = package_dir.file_with(u'packaged_file', fixture.file1_contents)
        another_file = package_dir.file_with(u'packaged_file2', fixture.file2_contents)

        pkg_resources.working_set.add(easter_egg)
        easter_egg.set_module_path(egg_dir.name)

        class MainUI(UserInterface):
            def assemble(self):
                to_concatenate = [PackagedFile(u'test==1.0', u'packaged_files', u'packaged_file'),
                                  PackagedFile(u'test==1.0', u'packaged_files', u'packaged_file2')]
                list_of_files = [ConcatenatedFile(fixture.filename, to_concatenate)]
                self.define_static_files(u'/files', list_of_files)

        fixture.config.reahlsystem.debug = False  # To enable minification 
        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # How the first file would be accessed
        browser.open('/files/%s' % fixture.filename)
        vassert( browser.raw_html == fixture.expected_result )

    @test(WebFixture)
    def file_download_details(self, fixture):
        """FileDownloadStub (the GET response for a StaticFileResource) works correctly in
          different scenarios of partial GETting too."""
        
        file_content = 'some content'
        server_file = temp_file_with(file_content, 'afile.css')
        @stubclass(FileDownload)
        class FileDownloadStub(FileDownload):
            chunk_size = 1
        response = FileDownloadStub(FileOnDisk(server_file.name, u'/path/for/the/file'))
        
        # Case: The whole content is sent, in chunk_size bits
        read = [i for i in response.app_iter]
        expected = [i for i in file_content]
        vassert( read == expected )

        # Case: Headers are set correctly
        vassert( response.content_type == 'text/css' )
        vassert( not response.content_encoding )
        vassert( response.content_length == len(file_content) )

        mtime = datetime.datetime.fromtimestamp(int(os.path.getmtime(server_file.name)))
        vassert( response.last_modified.replace(tzinfo=None) == mtime )
        tag_mtime, tag_size, tag_hash = response.etag.split('-')
        mtime = str(os.path.getmtime(server_file.name))
        vassert( tag_mtime == mtime )
        vassert( tag_size == str(len(file_content)) )
        vassert( tag_hash == str(abs(hash(server_file.name))) )
        
        # Case: conditional response is supported
        vassert( response.conditional_response )

        # Case: partial response is supported - different cases:
        #      - normal case
        actual = [i for i in response.app_iter.app_iter_range(3,7)]
        expected = [i for i in file_content[3:8]]
        vassert( actual == expected )

        #      - no end specified
        actual = [i for i in response.app_iter.app_iter_range(3)]
        expected = [i for i in file_content[3:]]
        vassert( actual == expected )

        #      - no start specified
        actual = [i for i in response.app_iter.app_iter_range(end=7)]
        expected = [i for i in file_content[:8]]
        vassert( actual == expected )

        #      - where the last chunk read would stretch past end
        response.chunk_size = 2
        actual = ''.join([i for i in response.app_iter.app_iter_range(end=6)])
        expected = file_content[:7]
        vassert( actual == expected )

        #      - where start > end
        response.chunk_size = 1
        actual = [i for i in response.app_iter.app_iter_range(start=7, end=3)]
        expected = ['']
        vassert( actual == expected )

        #      - where start < 0
        actual = [i for i in response.app_iter.app_iter_range(start=-10, end=7)]
        expected = [i for i in file_content[:8]]
        vassert( actual == expected )

        #      - where end > length of file
        actual = [i for i in response.app_iter.app_iter_range(start=3, end=2000)]
        expected = [i for i in file_content[3:]]
        vassert( actual == expected )

        #      - where start > length of file        
        actual = [i for i in response.app_iter.app_iter_range(start=700)]
        expected = ['']
        vassert( actual == expected )

    @test(WebFixture)
    def standard_reahl_files(self, fixture):
        """The framework creates certain static files by default."""
        wsgi_app = ReahlWSGIApplication(fixture.config)
        browser = Browser(wsgi_app)

        browser.open(u'/static/html5shiv-printshiv-3.6.3.js')
        vassert( browser.last_response.content_length > 0 )

        browser.open(u'/static/IE9.js')
        vassert( browser.last_response.content_length > 0 )

        browser.open(u'/static/reahl.js')
        vassert( browser.last_response.content_length > 0 )

        browser.open(u'/static/reahl.css')
        vassert( browser.last_response.content_length > 0 )

        browser.open(u'/static/runningon.png')
        vassert( browser.last_response.content_length > 0 )






# Copyright 2010, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
The interfaces for persisted classes that are needed by the core Reahl framework. Different
implementations of the framework can be provided by implementing these.
"""

from abc import ABCMeta, abstractmethod

from reahl.interfaces import UserSessionProtocol


class WebUserSessionProtocol(UserSessionProtocol):
    """The protocol for an implementation to a class used via `web.session_class`."""
    @classmethod
    @abstractmethod
    def get_or_create_session(cls): 
        """Returns a session, creating a new one if none can be determined. If one can be determined from
           the browser, that session is returned."""
        pass

    @abstractmethod
    def set_session_key(self, response): 
        """Called at the end of a request loop to enable an implementation to save some identifying information
           to the response (such as setting a cookie with the ID of the current session).
        """
        pass


class UserInputProtocol(object):
    """User input, typed as strings on a form is persisted using this class, for the current user's session
       for use in a subsequent request. Used via `web.persisted_userinput_class`.
    """
    __metaclass__ = ABCMeta

    @classmethod
    @abstractmethod
    def clear_for_form(cls, form):
        """Removes all the user input associated with the given :class:`reahl.web.ui.Form`."""
        pass

    @classmethod
    @abstractmethod
    def get_previously_entered_for_form(cls, form, input_name): 
        """Returns the user input associated with the given :class:`reahl.web.ui.Form`, as previously saved using
           `input_name` as name."""
        pass

    @classmethod
    @abstractmethod
    def save_input_value_for_form(cls, form, input_name, value):
        """Persists `value` as the value of the user input associated with the given :class:`reahl.web.ui.Form`,
           using `input_name` as name."""
        pass
        
    @abstractmethod
    def __eq__(self, other): 
        """Is required to be implemented."""
        pass

    @abstractmethod
    def __neq__(self, other): 
        """Is required to be implemented."""
        pass


class PersistedExceptionProtocol(UserInputProtocol):
    """When a :class:`reahl.component.exceptions.DomainException` happens during Form submission, the 
       exception is persisted using this class, for the current user's session for use in a subsequent 
       request. Used via `web.persisted_exception_class`.
    """
    @classmethod
    @abstractmethod
    def get_exception_for_form(self, form): 
        """Retrieves an exception previously saved for the given :class:`reahl.web.ui.Form`, or None if
           not found."""
        pass

    @classmethod
    @abstractmethod
    def get_exception_for_input(self, form, input_name): 
        """Retrieves an exception previously saved for the given :class:`reahl.web.ui.Form` and `input_name`
           or None if not found."""
        pass
    
    @classmethod
    @abstractmethod
    def clear_for_all_inputs(cls, form): 
        """Removes all saved Exceptions associated with the given :class:`reahl.web.ui.Form`."""
        pass

      
class PersistedFileProtocol(object):
    """When a file is uploaded, file is persisted using this class, for the current user's session 
       for use in a subsequent request. Used via `web.persisted_file_class`.
    """
    __metaclass__ = ABCMeta
    @property
    @abstractmethod
    def file_obj(self): 
        """Returns an object with traditional `.read` and `.seek` methods which can be used to
           read the contents of the persisted file.
        """
        pass

    @classmethod
    @abstractmethod
    def clear_for_form(cls, form): 
        """Removes all files previously saved for the given :class:`reahl.web.ui.Form`."""
        pass

    @abstractmethod
    def __eq__(self, other): 
        """Is required to be implemented."""
        pass

    @abstractmethod
    def __neq__(self, other): 
        """Is required to be implemented."""
        pass

    @classmethod
    @abstractmethod
    def get_persisted_for_form(cls, form, input_name): 
        """Returns the previously persisted file for the given :class:`reahl.web.ui.Form`,
           using the given `input_name` as name."""
        pass

    @classmethod
    @abstractmethod
    def add_persisted_for_form(cls, form, input_name, uploaded_file): 
        """Saves the given `uploaded_file` (a :class:`cgi.FileStorage`) using the given `input_name`
           for the given :class:`reahl.web.ui.Form`.
        """
        pass

    @classmethod
    @abstractmethod
    def remove_persisted_for_form(cls, form, input_name, filename): 
        """Removes previously persisted file with name `filename`, as saved for the given 
           :class:`reahl.web.ui.Form` and `input_name`."""
        pass

    @classmethod
    @abstractmethod
    def is_uploaded_for_form(cls, form, input_name, filename): 
        """Answers whether a file with name `filename` has previously been saved for the given 
           :class:`reahl.web.ui.Form` and `input_name`."""
        pass



# Copyright 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from nose.tools import istest
from reahl.tofu import Fixture, test
from reahl.tofu import vassert

from reahl.web_dev.fixtures import WebFixture
from reahl.web.fw import UserInterface, IdentityDictionary
from reahl.web.ui import TwoColumnPage
from reahl.webdev.tools import Browser
from reahl.component.i18n import Translator


from reahl.web_dev.fixtures import WebBasicsMixin
from reahl.web.fw import UrlBoundView, Url
from reahl.component.context import ExecutionContext
from reahl.webdev.tools import WidgetTester

class XXFixture(Fixture, WebBasicsMixin):
    def new_view(self):
        current_path = Url(ExecutionContext.get_context().request.url).path
        view = UrlBoundView(None, current_path, u'Harness view', {})
        return view

@istest
class XXTests(object):
    @test(XXFixture)
    def test(self, fixture):
        from reahl.web.ui import P

        p = P(fixture.view)
        tester = WidgetTester(p)
        rendered_html = tester.render_html()
        vassert( rendered_html == u'<p></p>' )


        
    
@test(WebFixture)
def i18n_urls(fixture):
    """The current locale is determined by reading the first segment of the path. If the locale is not present in the
    path, web.default_url_locale is used."""
    _ = Translator(u'reahl-web')

    class I18nUI(UserInterface):
        def assemble(self):
            view = self.define_view(u'/aview', title=_(u'A View'))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(TwoColumnPage)
            self.define_user_interface(u'/a_ui',  I18nUI,  IdentityDictionary(), name=u'test_ui')
            
    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open(u'/a_ui/aview')
    vassert( browser.title == u'A View' )

    browser.open(u'/af/a_ui/aview')
    vassert( browser.title == u'\'n Oogpunt' )

    fixture.context.config.web.default_url_locale = 'af'
    browser.open(u'/a_ui/aview')
    vassert( browser.title == u'\'n Oogpunt' )
    
    browser.open(u'/en_gb/a_ui/aview')
    vassert( browser.title == u'A View' )



# Copyright 2010-2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Tools for breaking long lists into shorter lists that can be paged.
"""

from abc import ABCMeta, abstractmethod, abstractproperty
from functools import partial

from reahl.component.i18n import Translator
from reahl.component.decorators import memoized
from reahl.component.modelinterface import IntegerField, exposed
from reahl.web.fw import WidgetResult, Bookmark
from reahl.web.ui import HMenu, A, Panel

_ = Translator(u'reahl-web')


class PageIndexProtocol(object):
    __metaclass__ = ABCMeta
    @abstractproperty
    def pages_in_range(self): pass
    @abstractproperty
    def current_page(self): pass
    @abstractproperty
    def start_page(self): pass
    @abstractproperty
    def end_page(self): pass
    @abstractproperty
    def previous_page(self): pass
    @abstractproperty
    def next_page(self): pass
    @abstractproperty
    def last_page(self): pass
    @abstractproperty
    def has_next_page(self): pass
    @abstractproperty
    def has_previous_page(self): pass


class PagedPanel(Panel):
    """A :class:`Panel` whose contents change, depending on the page selected by a user from a :class:`PageMenu`.
       A programmer should subclass from PagedPanel, supplying an `__init__` method which populates the PagedPanel
       with appropriate contents, based on its `.current_contents`.
       
       .. admonition:: Styling
          
          Represented in HTML by an <div> element.

       :param view: (See :class:`reahl.web.fw.Widget`)
       :param page_index: The :class:`PageIndex` to use to supply contents to the pages displayed by this PagedPanel.
       :param css_id: (See :class:`HTMLElement`)
    """
    def __init__(self, view, page_index, css_id):
        self.page_index = page_index
        super(PagedPanel, self).__init__(view, css_id=css_id)
        self.enable_refresh()
    
    @classmethod
    def get_bookmark(cls, description=None, page_number=1):
        return Bookmark(u'', u'', description=description or u'%s' % page_number,
                        query_arguments={u'current_page_number': page_number}, ajax=True)

    @exposed
    def query_fields(self, fields):
        fields.current_page_number = self.page_index.fields.current_page_number

    @property
    def current_contents(self):
        """The list of items that should be displayed for the current page."""
        return self.page_index.current_page.contents


class Page(object):
    def __init__(self, number, description, contents_getter):
        self.number = number
        self.description = description
        self.contents_getter = contents_getter

    @property
    def contents(self):
        return self.contents_getter()
    

class PageIndex(PageIndexProtocol):
    """An object responsible for breaking a long list of items up into shorter lists for display. Each such
       shorter list is referred to as a page. Different ways of breaking long lists into smaller lists
       are provided by subclasses.

       :param current_page_number: The initial page shown to users.
       :param start_page_number: The first page that should be listed in the current :class:`PageMenu`.
       :param max_page_links: The number of pages to be shown in the current :class:`PageMenu`.
    """
    def __init__(self, current_page_number=1, start_page_number=1, max_page_links=5):
        super(PageIndex, self).__init__()
        self.current_page_number = current_page_number
        self.start_page_number = start_page_number
        self.max_page_links = max_page_links
        
    @exposed
    def fields(self, fields):
        fields.current_page_number = IntegerField(required=False, default=1)
        fields.start_page_number =   IntegerField(required=False, default=1)

    @abstractmethod
    def get_contents_for_page(self, page_number): 
        """Override this method in subclasses to obtain the correct list of items for the given `page_number`."""
        pass

    @abstractproperty
    def total_number_of_pages(self): 
        """Override this @property in subclasses to state what the total number of pages is."""    
        pass
    
    def get_description_for_page(self, page_number):
        return unicode(page_number)
    
    @property
    @memoized
    def current_page(self):
        return self.get_page_number(self.current_page_number)
    
    def get_page_number(self, page_number):
        return Page(page_number, self.get_description_for_page(page_number), partial(self.get_contents_for_page, page_number))

    @property
    @memoized
    def pages_in_range(self):
        return [self.get_page_number(page_number) 
                for page_number in range(self.start_page_number, self.end_page.number+1)]

    @property
    @memoized
    def start_page(self):
        return self.get_page_number(self.start_page_number)

    @property
    @memoized
    def end_page(self):
        page_number = self.start_page_number+min(self.max_page_links-1, self.total_number_of_pages-(self.start_page_number));
        return self.get_page_number(page_number)
        
    @property
    @memoized
    def previous_page(self):
        page_number = max(1, self.start_page_number - self.max_page_links)
        return self.get_page_number(page_number)

    @property
    @memoized
    def next_page(self):
        page_number = min(self.end_page.number + 1, self.total_number_of_pages)
        return self.get_page_number(page_number)
                
    @property
    @memoized
    def last_page(self):
        page_number =  max(self.total_number_of_pages - self.max_page_links + 1, 1)
        return self.get_page_number(page_number)

    @property
    @memoized
    def has_next_page(self):
        return self.end_page.number < self.total_number_of_pages

    @property
    @memoized
    def has_previous_page(self):
        return self.start_page.number > 1


class PageMenu(HMenu):
    """An HMenu, which lists the pages of items that can be navigated by a user. If there are
       many pages, only a small subset is shown, with controls allowing the user to browse to
       the wanted page number and choose it.

       .. admonition:: Styling
       
          Rendered as a <ul class="reahl-menu reahl-horizontal reahl-pagemenu">. The <a> inside its
          first <li> has class="first", the <a> in the second <li> has class="prev". The <a> in the 
          second last <li> has class="next", and the <a> in the last <li> has class="last".
          

       :param view: (See :class:`reahl.web.fw.Widget`)
       :param css_id: (See :class:`HTMLElement`)
       :param page_index: The :class:`PageIndex` whose pages are displayed by this PageMenu.
       :param page_container: The :class:`PagedPanel` in which the contents of a page is displayed.
    """
    def __init__(self, view, css_id, page_index, paged_panel):
        self.page_index = page_index
        super(PageMenu, self).__init__(view, [], css_id=css_id)
        self.append_class(u'reahl-pagemenu')

        self.paged_panel = paged_panel
        self.add_items()
        self.enable_refresh()

    def add_items(self):
        links = []

        first = A.from_bookmark(self.view, self.get_bookmark(start_page_number=1, description=u'|<'))
        first.append_class(u'first')
        first.set_active(self.page_index.has_previous_page)
        previous = A.from_bookmark(self.view, self.get_bookmark(start_page_number=self.page_index.previous_page.number, description=u'<'))
        previous.append_class(u'prev')
        previous.set_active(self.page_index.has_previous_page)
        links.extend([first, previous])

        for page in self.page_index.pages_in_range:
            link = A.from_bookmark(self.view, self.paged_panel.get_bookmark(page_number=page.number, description=page.description))
            links.append(link)

        next = A.from_bookmark(self.view, self.get_bookmark(start_page_number=self.page_index.next_page.number, description=u'>'))
        next.append_class(u'next')
        next.set_active(self.page_index.has_next_page)
        last = A.from_bookmark(self.view, self.get_bookmark(start_page_number=self.page_index.last_page.number, description=u'>|'))
        last.append_class(u'last')
        last.set_active(self.page_index.has_next_page)
        links.extend([next, last])

        self.set_items_from(links)

    @classmethod
    def get_bookmark(self, description=None, start_page_number=1):
        return Bookmark.for_widget(description=description or u'%s' % start_page_number,
                                   query_arguments={u'start_page_number': start_page_number})

    @exposed
    def query_fields(self, fields):
        fields.start_page_number = self.page_index.fields.start_page_number
    

class SequentialPageIndex(PageIndex):
    """A PageIndex that breaks a list of items up into smaller lists, by cutting the original list
       into sections that have a maximum number of items per page.
       
       :param items: The long list of items.
       :param items_per_page: The maximum number of items to allow on a page.
       :param current_page_number: (See :class:`PageIndex`)
       :param start_page_number: (See :class:`PageIndex`)
       :param max_page_links: (See :class:`PageIndex`)
    """
    def __init__(self, items, items_per_page=5, current_page_number=1, start_page_number=1, max_page_links=4):
        super(SequentialPageIndex, self).__init__(current_page_number=current_page_number,
                                              start_page_number=start_page_number,
                                              max_page_links=max_page_links)
        self.items = items
        self.items_per_page = items_per_page

    def get_contents_for_page(self, page_number):
        range_start = (page_number-1)*self.items_per_page
        range_end = range_start+min(self.items_per_page, len(self.items)-(range_start))-1
        return self.items[range_start:range_end+1]

    @property
    @memoized
    def total_number_of_pages(self):
        return ((len(self.items)-1) // (self.items_per_page))+1


class AnnualItemOrganiserProtocol(object):
    """Manages a list of items, each of which is seen to be for a particular year.
    """
    __metaclass__ = ABCMeta
    @abstractmethod
    def get_years(self): 
        """Returns a list of integers, each representing a year which is applicable to at least one item in a list of items."""
        pass

    @abstractmethod
    def get_items_for_year(self, year): 
        """Returns a list if items to which `year` (an integer) is applicable."""
        pass


class AnnualPageIndex(PageIndex):
    """A PageIndex that breaks a list of items up into smaller lists, by arranging all items
       that have the same year on the same page. 
       
       :param annual_item_organiser: An object that implements :class:`AnnualItemOrganiserProtocol`. Its methods
                                     will be called to find the relevent items, or determine what years are applicable.
       :param current_page_number: (See :class:`PageIndex`)
       :param start_page_number: (See :class:`PageIndex`)
       :param max_page_links: (See :class:`PageIndex`)
    """
    def __init__(self, annual_item_organiser, current_page_number=1, start_page_number=1, max_page_links=4):
        super(AnnualPageIndex, self).__init__(current_page_number=current_page_number,
                                          start_page_number=start_page_number,
                                          max_page_links=max_page_links)
        self.annual_item_organiser = annual_item_organiser

    def get_contents_for_page(self, page_number):
        year = self.years[page_number-1]
        return self.annual_item_organiser.get_items_for_year(year)

    def get_description_for_page(self, page_number):
        return unicode(self.years[page_number-1])

    @property
    def years(self):
        return self.annual_item_organiser.get_years()

    @property
    @memoized
    def total_number_of_pages(self):
        return len(self.years)





# Copyright 2009-2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from webob import Request
from webob.exc import HTTPSeeOther

from nose.tools import istest
from reahl.tofu import Fixture, test, scenario
from reahl.tofu import vassert, expected
from reahl.stubble import stubclass, EmptyStub, exempt

from reahl.web.ui import Input, TextInput, Button, Form, ValidationException, \
                          LabelOverInput, CueInput, CheckboxInput, TextInput, InputLabel, ButtonInput,\
                          PasswordInput, Button, LabelledInlineInput, LabelledBlockInput, P,\
                          TextArea, SelectInput, RadioButtonInput

from reahl.web.fw import WebExecutionContext
from reahl.component.modelinterface import Field, EmailField, BooleanField,\
                             RequiredConstraint, MinLengthConstraint, PatternConstraint, RemoteConstraint,\
                             Event, Allowed, exposed, Action, Choice, ChoiceGroup, ChoiceField, IntegerField,\
                             MultiChoiceField, DateField
from reahl.component.exceptions import IsInstance
from reahl.web_dev.fixtures import WebFixture
from reahl.webdev.tools import WidgetTester, Browser, XPath
from reahl.component_dev.fieldtests import FieldMixin

from reahl.partymodel import Party

class InputMixin(FieldMixin):
    def new_form(self):
        return Form(self.view, u'test')

    def new_event(self, label=u'click me'):
        event = Event(label=label)
        event.bind(u'aname', None)
        self.form.define_event_handler(event)
        return event

class SimpleInputFixture(WebFixture, InputMixin):
    pass

class InputMixin2(InputMixin):
    def new_field(self, label=u'the label'):
        return Field(label=label)
    def new_model_object(self):
        fixture = self
        class ModelObject(object):
            def handle_event(self):
                pass
            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me', action=Action(self.handle_event))
            @exposed
            def fields(self, fields):
                fields.an_attribute = fixture.field
        return ModelObject()

class InputStateFixture(WebFixture, InputMixin):
    def new_field(self):
        field = EmailField(default=u'default value')
        field.bind(u'an_attribute', self.model_object)
        return field

    def new_input(self):    
        return Input(self.form, self.field)

    @scenario
    def defaulted_input_no_value_on_domain(self):
        del self.model_object.an_attribute
        self.input.prepare_input()

        self.expected_state = u'defaulted'
        self.expected_value = u'default value'

    @scenario
    def defaulted_input_value_on_domain(self):
        self.input.prepare_input()
        self.model_object.an_attribute = u'something'

        self.expected_state = u'defaulted'
        self.expected_value = u'something'

    @scenario
    def validly_entered(self):
        self.input.enter_value(u'someone@home.org')
        self.input.prepare_input()

        self.expected_state = u'validly_entered' 
        self.expected_value = u'someone@home.org'

    @scenario
    def empty_input_non_required_field(self):
        self.input.enter_value(u'')
        self.input.prepare_input()

        self.expected_state = u'validly_entered' 
        self.expected_value = u''

    @scenario
    def empty_input_required_field(self):
        self.field.make_required(u'')
        self.input.enter_value(u'')
        self.input.prepare_input()

        self.expected_state = u'invalidly_entered' 
        self.expected_value = u''

    @scenario
    def invalidly_entered(self):
        self.input.enter_value(u'not an email address')
        self.input.prepare_input()

        self.expected_state = u'invalidly_entered' 
        self.expected_value = u'not an email address'



@istest
class InputDetailsTests(object):
    @test(InputStateFixture)
    def the_states_of_an_input(self, fixture):
        """An Input can be in one of three states: defaulted, invalidly_entered or validly_entered. Depending
           on the state of the input, it will have a different value when rendering.

           The states:
             - defaulted: the user has not entered anything
               value:     the value on the model object which is bound to its Field if that object
                          has such an an_attribute, else the default value of the Field.
             - invalidly_entered: the user entered a value which failed validation
               value:             the invalid value which the user entered
             - validly_entered: the user entered a valid value
               value:           the value which the user entered even if the underlying 
                                model object has a corresponding value
        """
        vassert( fixture.input.get_input_status() == fixture.expected_state )
        vassert( fixture.input.value == fixture.expected_value )



@istest
class BasicInputTests(object):
    class Scenarios(WebFixture, InputMixin):
        @scenario
        def input_label(self):
            html_input = TextInput(self.form, self.field)
            self.widget = InputLabel(html_input)
            self.expected_html = u'<label for="an_attribute">the label</label>'
            self.field_controls_visibility = True

        @scenario
        def input_label_with_text(self):
            html_input = TextInput(self.form, self.field)
            self.widget = InputLabel(html_input, text=u'some text')
            self.expected_html = u'<label for="an_attribute">some text</label>'
            self.field_controls_visibility = True

        @scenario
        def button_input(self):
            self.widget = self.form.add_child(ButtonInput(self.form, self.event))
            self.expected_html = u'<input name="event.aname?" form="test" type="submit" value="click me">'
            self.field_controls_visibility = False

        @scenario
        def button(self):
            self.widget = self.form.add_child(Button(self.form, self.event))
            self.expected_html = u'<span class="reahl-button"><input name="event.aname?" form="test" type="submit" value="click me"></span>'
            self.field_controls_visibility = False

        @scenario
        def labelled_inline_input(self):
            self.model_object.an_attribute = u'field value'
            self.widget = self.form.add_child(LabelledInlineInput(TextInput(self.form, self.field)))
            self.expected_html = u'<span class="reahl-labelledinput">'\
                                 u'<label for="an_attribute">the label</label>' \
                                 u'<span><input name="an_attribute" form="test" type="text" value="field value" class="reahl-textinput"></span>' \
                                 u'</span>'
            self.field_controls_visibility = True

        @scenario
        def labelled_block_input(self):
            self.model_object.an_attribute = u'field value'
            self.widget = self.form.add_child(LabelledBlockInput(TextInput(self.form, self.field)))
            self.expected_html = u'<div class="reahl-labelledinput yui-g">'\
                                 u'<div class="first yui-u"><label for="an_attribute">the label</label></div>' \
                                 u'<div class="yui-u"><input name="an_attribute" form="test" type="text" value="field value" class="reahl-textinput"></div>' \
                                 u'</div>'
            self.field_controls_visibility = True

        @scenario
        def password(self):
            self.widget = self.form.add_child(PasswordInput(self.form, self.field)) 
            self.expected_html = u'<input name="an_attribute" form="test" type="password">'
            self.field_controls_visibility = True

        @scenario
        def checkbox_true(self):
            """A checkbox needs a 'checked' an_attribute if its field is True. It also renders ONLY the
                validation message for its required validation_constraint, not for all constraints"""
            self.model_object.an_attribute = True

            self.field = BooleanField(required=True, label=u'my text', required_message=u'$label is needed here')
            self.field.bind(u'an_attribute', self.model_object)

            self.widget = self.form.add_child(CheckboxInput(self.form, self.field))
            self.expected_html = u'<input name="an_attribute" checked="checked" form="test" required="*" type="checkbox" ' \
                                 u'class="{&quot;validate&quot;: {&quot;messages&quot;: {&quot;required&quot;: &quot;my text is needed here&quot;}}}">'
            self.field_controls_visibility = True

        @scenario
        def checkbox_false(self):
            self.checkbox_true()
            self.model_object.an_attribute = False
            self.expected_html = u'<input name="an_attribute" form="test" required="*" type="checkbox" '\
                                 u'class="{&quot;validate&quot;: {&quot;messages&quot;: {&quot;required&quot;: &quot;my text is needed here&quot;}}}">'
            self.field_controls_visibility = True

        @scenario
        def text_area_input(self):
            self.widget = self.form.add_child(TextArea(self.form, self.field, rows=30, columns=20))
            self.expected_html = u'<textarea name="an_attribute" cols="20" rows="30">field value</textarea>'
            self.field_controls_visibility = True

        @scenario
        def select_input(self):
            self.model_object.an_attribute = 2

            choices = [Choice(False, BooleanField(label=u'None')),
                       ChoiceGroup(u'grouped', [
                           Choice(1, IntegerField(label=u'One')), 
                           Choice(2, IntegerField(label=u'Two'))])
                      ]
            self.field = ChoiceField(choices)
            self.field.bind(u'an_attribute', self.model_object)

            self.widget = self.form.add_child(SelectInput(self.form, self.field))
            group = u'<optgroup label="grouped"><option value="1">One</option><option selected="selected" value="2">Two</option></optgroup>'
            option = u'<option value="off">None</option>'
            self.expected_html = u'<select name="an_attribute" form="test">%s%s</select>' % (option, group)
            self.field_controls_visibility = True

        @scenario
        def select_input_multi(self):
            self.model_object.an_attribute = [2]

            choices = [Choice(1, IntegerField(label=u'One')), 
                       Choice(2, IntegerField(label=u'Two'))
                      ]
            self.field = MultiChoiceField(choices)
            self.field.bind(u'an_attribute', self.model_object)

            self.widget = self.form.add_child(SelectInput(self.form, self.field)) 
            options = u'<option value="1">One</option><option selected="selected" value="2">Two</option>'
            self.expected_html = u'<select name="an_attribute" form="test" multiple="multiple">%s</select>' % (options)
            self.field_controls_visibility = True

        @scenario
        def radio_button(self):
            self.model_object.an_attribute = 2

            choices = [ Choice(1, IntegerField(label=u'One')), 
                        Choice(2, IntegerField(label=u'Two')) ]
            self.field = ChoiceField(choices)
            self.field.bind(u'an_attribute', self.model_object)

            self.widget = self.form.add_child(RadioButtonInput(self.form, self.field))

            radio_button = u'<span class="reahl-radio-button"><input name="an_attribute"%s form="test" type="radio" value="%s">%s</span>'

            outer_div = u'<div class="reahl-radio-button-input">%s</div>'
            buttons = (radio_button % (u'', u'1', u'One')) + (radio_button % (u' checked="checked"', u'2', u'Two'))
            self.expected_html = outer_div % buttons
            self.field_controls_visibility = True

    @test(Scenarios)
    def basic_rendering(self, fixture):
        """What the rendered html for a number of simple inputs look like."""
        
        tester = WidgetTester(fixture.widget)

        # Case: Normal behaviour
        actual = tester.render_html()
        vassert( actual == fixture.expected_html )

        # Case: When the Widget should not be visible
        fixture.field.access_rights.readable = Allowed(False)
        fixture.field.access_rights.writable = Allowed(False)
        actual = tester.render_html()
        if fixture.field_controls_visibility:
            vassert( actual == u'' )
        else:
            vassert( actual == fixture.expected_html )
        

    @test(SimpleInputFixture)
    def wrong_args_to_input(self, fixture):
        """Passing the wrong arguments upon constructing an Input results in an error."""
        
        with expected(IsInstance):
            Input(fixture.form, EmptyStub()) 

        with expected(IsInstance):
            Input(EmptyStub(), Field()) 


    class CheckboxFixture(WebFixture, InputMixin2):
        def new_field(self):
            return BooleanField(label=u'my text')

    @test(CheckboxFixture)
    def marshalling_of_checkbox(self, fixture):
        """When a form is submitted, the value of a checkbox is derived from
           whether the checkbox is included in the submission or not."""

        model_object = fixture.model_object
        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                checkbox = self.add_child(CheckboxInput(self, model_object.fields.an_attribute))
                self.define_event_handler(model_object.events.an_event)
                self.add_child(Button(self, model_object.events.an_event))
                fixture.checkbox = checkbox

        wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory(u'myform'))
        fixture.reahl_server.set_app(wsgi_app)
        fixture.driver_browser.open('/')

        # Case: checkbox is submitted with form (ie checked)
        fixture.driver_browser.check(u"//input[@type='checkbox']")
        fixture.driver_browser.click(u"//input[@value='click me']")

        vassert( model_object.an_attribute )
        vassert( fixture.checkbox.value == u'on' )        

        # Case: checkbox is not submitted with form (ie unchecked)
        fixture.driver_browser.uncheck(u"//input[@type='checkbox']")
        fixture.driver_browser.click(u"//input[@value='click me']")

        vassert( not model_object.an_attribute )
        vassert( fixture.checkbox.value == u'off' )        



class LabelOverInputFixture(WebFixture, InputMixin2):
    label_xpath = u"//form/span[@class='reahl-labelledinput reahl-labeloverinput']/label[@for='an_attribute' and text()='the label']"

@istest
class LabelOverInputTests(object):
    @test(LabelOverInputFixture)
    def label_behaviour(self, fixture):
        """The label is displayed over the input, but it disappears when selected
           or when there's a value in the input."""
        
        model_object = fixture.model_object

        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                self.add_child(LabelOverInput(TextInput(self, model_object.fields.an_attribute)))
                self.define_event_handler(model_object.events.an_event)
                self.add_child(Button(self, model_object.events.an_event))

        wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory(u'myform'), enable_js=True)
        fixture.reahl_server.set_app(wsgi_app)

        # Case: the field is empty and does not have a default value
        # - render initially with label
        fixture.driver_browser.open('/')
        vassert( fixture.driver_browser.is_element_present(fixture.label_xpath) )
        fixture.driver_browser.wait_for_element_visible(fixture.label_xpath)
        
        # - if click, label disappears, input gets focus
        fixture.driver_browser.click(fixture.label_xpath)
        fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)

        # - if focus lost, label appears again if nothing is entered
        fixture.driver_browser.press_tab(u'//input')
        fixture.driver_browser.wait_for_element_visible(fixture.label_xpath)
        
        # - if focus lost, label does not appear again if something is entered
        fixture.driver_browser.type(u'//input[@type="text"]', 'my@email.com')
        fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)
        fixture.driver_browser.press_tab(u'//input')
        fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)

        
        # Case: the field has a defaulted value
        model_object.an_attribute = 'some value'
        fixture.driver_browser.open('/')
        
        # - render initially without a label
        fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)

        # - if value removed, and focus lost, label appears
        fixture.driver_browser.type(u'//input[@type="text"]', '')
        fixture.driver_browser.press_tab(u'//input')
        fixture.driver_browser.wait_for_element_visible(fixture.label_xpath)

    @test(LabelOverInputFixture)
    def basic_rendering(self, fixture):
        """What the html for a LabelOverInput looks like."""

        fixture.field.bind(u'an_attribute', fixture.model_object)
        label_over_input = fixture.form.add_child(LabelOverInput(TextInput(fixture.form, fixture.field))) 
        tester = WidgetTester(label_over_input)

        # Case: HTML - with a value
        #  - has CSS class reahl-labeloverinput
        #  - the label is hidden
        fixture.model_object.an_attribute = u'my value'
        actual = tester.render_html()
        expected = u'<span class="reahl-labelledinput reahl-labeloverinput">'\
                   u'<label for="an_attribute" hidden="true">the label</label>' \
                   u'<span><input name="an_attribute" form="test" type="text" value="my value" class="reahl-textinput"></span>'\
                   u'</span>'
        vassert( actual == expected )

        # Case: HTML - without a value
        #  - has CSS class reahl-labeloverinput
        #  - the label is not hidden
        del fixture.model_object.an_attribute
        actual = tester.render_html()
        expected = u'<span class="reahl-labelledinput reahl-labeloverinput">'\
                   u'<label for="an_attribute">the label</label>' \
                   u'<span><input name="an_attribute" form="test" type="text" value="" class="reahl-textinput"></span>'\
                   u'</span>'
        vassert( actual == expected )

        # Case: when the underlying input is not visible
        fixture.field.access_rights.readable = Allowed(False)
        fixture.field.access_rights.writable = Allowed(False)
        actual = tester.render_html()
        vassert( actual == u'' )

class CueInputFixture(WebFixture, InputMixin2):
    cue_xpath = u"//div[@class='reahl-cueinput reahl-labelledinput yui-g']/div/div[@class='reahl-cue yui-u']/p"

    def new_text_input(self):
        return TextInput(self.form, self.field)

    def new_cue(self):
        return P(self.view, text=u'this is your cue')

    def new_cue_input(self):
        return CueInput(self.text_input, self.cue)

        
@istest
class CueInputTests(object):
    @test(CueInputFixture)
    def cue_behaviour(self, fixture):
        """The Cue appears when its input has focus only."""
        model_object = fixture.model_object

        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                self.add_child(CueInput(TextInput(self, model_object.fields.an_attribute), P(view, text=u'this is your cue')))
                self.define_event_handler(model_object.events.an_event)
                self.add_child(Button(self, model_object.events.an_event))

        wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory(u'myform'), enable_js=True)
        fixture.reahl_server.set_app(wsgi_app)
        fixture.driver_browser.open('/')

        # - initially rendered without cue
        vassert( fixture.driver_browser.is_element_present(fixture.cue_xpath) )
        fixture.driver_browser.wait_for_element_not_visible(fixture.cue_xpath)
        
        # - if click, cue appears
        fixture.driver_browser.click(u'//input[@type="text"]')
        fixture.driver_browser.wait_for_element_visible(fixture.cue_xpath)

        # - if focus lost, cue disappears again 
        fixture.driver_browser.press_tab('//input')
        fixture.driver_browser.wait_for_element_not_visible(fixture.cue_xpath)

    @test(CueInputFixture)
    def basic_rendering(self, fixture):
        """What the html for a CueInput looks like."""
        fixture.field.bind(u'an_attribute', fixture.model_object)
        cue = fixture.cue
        cue_input = fixture.cue_input
        fixture.form.add_child(cue_input) 
        fixture.model_object.an_attribute = u'my value'
 
        tester = WidgetTester(cue_input)

        # Case: normal behaviour
        actual = tester.render_html()
        expected = u'<div class="reahl-cueinput reahl-labelledinput yui-g">'\
                    u'<div class="first yui-u"><label for="an_attribute">the label</label></div>' \
                    u'<div class="yui-g">' \
                     u'<div class="first yui-u"><input name="an_attribute" form="test" type="text" value="my value" class="reahl-textinput"></div>' \
                     u'<div class="reahl-cue yui-u">' \
                      u'<p hidden="true">this is your cue</p>' \
                     u'</div>' \
                    u'</div>' \
                   u'</div>'
        vassert( actual == expected )

        # Case: when the underlying input is not visible
        fixture.field.access_rights.readable = Allowed(False)
        fixture.field.access_rights.writable = Allowed(False)
        actual = tester.render_html()
        vassert( actual == u'' )


class FuzzyTextInputFixture(WebFixture, InputMixin2):
    def new_field(self, label=u'the label'):
        return DateField(label=label)



@istest
class TextInputTests(object):
    @test(FuzzyTextInputFixture)
    def fuzzy(self, fixture):
        """A TextInput can be created as fuzzy=True. Doing this results in the possibly imprecise
           input that was typed by the user to be interpreted server-side and changed to the
           more exact representation in the client browser.  This happens upon the input losing focus."""
        model_object = fixture.model_object
        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                self.add_child(LabelledBlockInput(TextInput(self, model_object.fields.an_attribute, fuzzy=True)))
                self.define_event_handler(model_object.events.an_event)
                self.add_child(Button(self, model_object.events.an_event))

        wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory(u'myform'), enable_js=True)
        fixture.reahl_server.set_app(wsgi_app)
        browser = fixture.driver_browser
        browser.open('/')

        browser.type(XPath.input_labelled(u'the label'), u'20 November 2012')
        browser.press_tab(XPath.input_labelled(u'the label'))
        browser.wait_for(browser.is_element_value, XPath.input_labelled(u'the label'), u'20 Nov 2012')








# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from nose.tools import istest
from reahl.tofu import Fixture, test, scenario
from reahl.tofu import vassert, expected
from reahl.stubble import EmptyStub

from reahl.web.fw import Url
from reahl.web.ui import *
from reahl.webdev.tools import WidgetTester
from reahl.web_dev.fixtures import WebFixture

@istest
class HTMLWidgetBasics(object):
    @test(WebFixture)
    def basic_fixed_attributes(self, fixture):
        """How the static attributes of a Widget can be manipulated, queried and rendered."""

        widget = HTMLElement(fixture.view, u'x')
        tester = WidgetTester(widget)

        # Adding / setting
        widget.set_attribute(u'attr1', u'value1')
        widget.add_to_attribute(u'listattr', [u'one', u'two'])
        widget.add_to_attribute(u'listattr', [u'three'])
        
        # Querying
        vassert( widget.has_attribute(u'attr1') )
        vassert( widget.has_attribute(u'listattr') )
        vassert( not widget.has_attribute(u'notthere') )
        vassert( widget.attributes.v[u'attr1'] == u'value1' )
        vassert( widget.attributes.v[u'listattr'] == u'one three two' )
        
        # Rendering
        rendered = tester.render_html()
        vassert( rendered == u'<x attr1="value1" listattr="one three two">' )

        # Rendering - the order of attributes in the rendered output
        widget.set_attribute(u'id', u'123')
        widget.add_to_attribute(u'class', [u'z', u'b'])
        rendered = tester.render_html()
        vassert( rendered == u'<x id="123" attr1="value1" listattr="one three two" class="b z">' )


    @test(WebFixture)
    def handy_methods(self, fixture):
        """Some handy methods for special attributes."""

        widget = HTMLElement(fixture.view, u'x')
        tester = WidgetTester(widget)

        widget.set_title(u'the title')
        widget.set_id(u'the id')
        widget.append_class(u'two')
        widget.append_class(u'one')

        rendered = tester.render_html()
        vassert( rendered == u'<x id="the id" title="the title" class="one two">' )

    @test(WebFixture)
    def dynamically_determining_attributes(self, fixture):
        """A Widget can determine its attribute values at the latest possible stage, based on changing data."""

        class WidgetWithDynamicAttributes(HTMLElement):
            state = u'1'
            @property
            def attributes(self):
                attributes = super(WidgetWithDynamicAttributes, self).attributes
                attributes.set_to(u'dynamic', self.state)
                attributes.add_to(u'dynamiclist', [self.state])
                attributes.add_to(u'not-there', [u'a value'])
                attributes.remove_from(u'not-there', [u'a value'])
                return attributes

        widget = WidgetWithDynamicAttributes(fixture.view, u'x')
        widget.set_attribute(u'fixed', u'value1')
        tester = WidgetTester(widget)

        widget.state = u'1'
        rendered = tester.render_html()
        vassert( rendered == u'<x dynamic="1" dynamiclist="1" fixed="value1">' )

        widget.state = u'2'
        rendered = tester.render_html()
        vassert( rendered == u'<x dynamic="2" dynamiclist="2" fixed="value1">' )


    @test(WebFixture)
    def wrapper_widgets(self, fixture):
        """Sometimes, a HTMLElement is "wrapped" by an Input which represents the HTMLElement. In this case,
           dynamic as well as static attributes for the wrapped HTMLElement are obtained from its Input wrapper."""

        class WrapperWidget(Input):
            def get_wrapped_html_attributes(self, attributes):
                attributes.set_to(u'set-by-wrapper', u'rhythm and poetry')
                return attributes

        field = Field()
        field.bind('aname', field)
        wrapper = WrapperWidget(Form(fixture.view, u'formname'), field)
        widget = HTMLElement(fixture.view, u'x', wrapper_widget=wrapper)
        tester = WidgetTester(widget)

        # Case: dynamic attributes are supplied by the wrapper
        rendered = tester.render_html()
        vassert( rendered == u'<x set-by-wrapper="rhythm and poetry">' )

        # Case:several methods to set static attributes are delegated from the wrapper
        wrapper.set_id(u'myid')
        wrapper.set_title(u'mytitle')
        wrapper.add_to_attribute(u'list-attribute', [u'one', u'two'])
        wrapper.set_attribute(u'an-attribute', u'a value')

        rendered = tester.render_html()
        vassert( rendered == u'<x id="myid" an-attribute="a value" list-attribute="one two" set-by-wrapper="rhythm and poetry" title="mytitle">' )

    @test(WebFixture)
    def all_html_widgets_have_css_ids(self, fixture):
        """A Widget (for HTML) can have a css_id. If accessed, but not set, a ProgrammerError is raised."""
        
        widget = HTMLElement(fixture.view, u'x')
        widget.tag_name = u'x'
        tester = WidgetTester(widget)

        widget._css_id = None
        with expected(ProgrammerError): 
            widget.css_id
        
        widget._css_id = u'myid'
        vassert( widget.css_id == u'myid' )

    @test(WebFixture)
    def jquery_support(self, fixture):
        """Each HTML Widget has a jquery_selector that targets it uniquely. By default
           this makes use of the css_id of the Widget, which thus need to be set, however
           subclasses may override this behaviour. The jquery_selector (or a more lax 
           selector for the Widget) can also be further narrowed to a specific
           jquery selector context."""
        
        widget = HTMLElement(fixture.view, u'x')
        tester = WidgetTester(widget)

        with expected(ProgrammerError):  # css id not set
            widget.jquery_selector

        widget.set_id(u'anid')
        vassert( widget.jquery_selector == u'"#%s"' % widget.css_id )
        vassert( widget.attributes[u'id'].as_html_value() == widget.css_id )
        
        contextualised = widget.contextualise_selector(u'selector', u'context')
        vassert( contextualised ==  u'selector, "context"' )

        contextualised = widget.contextualise_selector(u'selector', None)
        vassert( contextualised ==  u'selector' )

    @test(WebFixture)
    def single_tags(self, fixture):
        """Definition of a HTMLElement."""
        
        single_tag = HTMLElement(fixture.view, u'single')
        tester = WidgetTester(single_tag)

        with expected(AssertionError):
            single_tag.add_child(P(fixture.view))

        rendered = tester.render_html()
        vassert( rendered == u'<single>' )

    @test(WebFixture)
    def closing_tags(self, fixture):
        """Definition of a HTMLElement with children."""
        
        closing_tag = HTMLElement(fixture.view, u'closing', children_allowed=True)
        tester = WidgetTester(closing_tag)

        closing_tag.add_child(P(fixture.view))

        rendered = tester.render_html()
        vassert( rendered == u'<closing><p></p></closing>' )



@istest
class BasicHTMLWidgets(object):
    class Scenarios(WebFixture):
        @scenario
        def text_node(self):
            self.widget = TextNode(self.view, u'text')
            self.expected_html = u'text'

        @scenario
        def a1(self):
            self.widget = A(self.view, Url(u'/xyz'))
            self.expected_html = u'<a href="/xyz"></a>'
            
        @scenario
        def a2(self):
            self.widget = A(self.view, Url(u'/xyz'), description=u'description')
            self.expected_html = u'<a href="/xyz">description</a>'

        @scenario
        def a3(self):
            def disallowed(): return False
            self.widget = A(self.view, Url(u'/xyz'), description=u'description', write_check=disallowed)
            self.expected_html = u'<a>description</a>'
            
        @scenario
        def h(self):
            self.widget = H(self.view, 2)
            self.expected_html = u'<h2></h2>'
            
        @scenario
        def p1(self):
            self.widget = P(self.view)
            self.expected_html = u'<p></p>'

        @scenario
        def p2(self):
            self.widget = P(self.view, text=u'text')
            self.expected_html = u'<p>text</p>'

        @scenario
        def p_with_slots(self):
            template_p = P(self.view, text=u'the {0} {{brown}} {slot2} jumps')
            self.widget = template_p.format(Span(self.view, text=u'quick'), slot2=Span(self.view, text=u'fox'))
            self.expected_html = u'<p>the <span>quick</span> {brown} <span>fox</span> jumps</p>'
            
        @scenario
        def title(self):
            self.widget = Title(self.view, u'text')
            self.expected_html = u'<title>text</title>'
            
        @scenario
        def link(self):
            self.widget = Link(self.view, u'rr', u'hh', u'tt')
            self.expected_html = u'<link href="hh" rel="rr" type="tt">'
            
        @scenario
        def header(self):
            self.widget = Header(self.view)
            self.expected_html = u'<header></header>'
            
        @scenario
        def footer(self):
            self.widget = Footer(self.view)
            self.expected_html = u'<footer></footer>'
            
        @scenario
        def li(self):
            self.widget = Li(self.view)
            self.expected_html = u'<li></li>'
            
        @scenario
        def ul(self):
            self.widget = Ul(self.view)
            self.expected_html = u'<ul></ul>'
            
        @scenario
        def img1(self):
            self.widget = Img(self.view, u'ss')
            self.expected_html = u'<img src="ss">'
            
        @scenario
        def img2(self):
            self.widget = Img(self.view, u'ss', alt=u'aa')
            self.expected_html = u'<img alt="aa" src="ss">'
            
        @scenario
        def span(self):
            self.widget = Span(self.view)
            self.expected_html = u'<span></span>'
            
        @scenario
        def span_with_text(self):
            self.widget = Span(self.view, text=u'some text')
            self.expected_html = u'<span>some text</span>'
            
        @scenario
        def div(self):
            self.widget = Div(self.view)
            self.expected_html = u'<div></div>'
            
        @scenario
        def nav(self):
            self.widget = Nav(self.view)
            self.expected_html = u'<nav></nav>'
            
        @scenario
        def article(self):
            self.widget = Article(self.view)
            self.expected_html = u'<article></article>'
            
        @scenario
        def label1(self):
            self.widget = Label(self.view)
            self.expected_html = u'<label></label>'
            
        @scenario
        def label2(self):
            self.widget = Label(self.view, text=u'text')
            self.expected_html = u'<label>text</label>'
            
        @scenario
        def fieldset1(self):
            self.widget = FieldSet(self.view)
            self.expected_html = u'<fieldset></fieldset>'
            
        @scenario
        def fieldset2(self):
            self.widget = FieldSet(self.view, label_text=u'text')
            self.expected_html = u'<fieldset><label>text</label></fieldset>'

    @test(Scenarios)
    def basic_html_widgets(self, fixture):
        """Several basic widgets merely correspond to html elements."""
        
        tester = WidgetTester(fixture.widget)
        rendered_html = tester.render_html()
        vassert( rendered_html == fixture.expected_html )

    @test(WebFixture)
    def view_rights_propagate_to_a(self, fixture):
        """The access rights specified for a View are propagated to an A, made from a Bookmark to that View."""
        fixture.view.write_check = EmptyStub()
        fixture.view.read_check = EmptyStub()
        a = A.from_bookmark(fixture.view, fixture.view.as_bookmark())
        vassert( a.read_check is fixture.view.read_check )
        vassert( a.write_check is fixture.view.write_check )
    
    @test(WebFixture)
    def text_node_can_vary(self, fixture):
        """A TextNode can vary its text if constructed with a getter for the value instead of a hardcoded value."""
        def getter():
            return fixture.current_value
        widget = TextNode(fixture.view, getter)
        tester = WidgetTester(widget)
        
        fixture.current_value = u'stuff'
        rendered = tester.render_html()
        vassert( rendered == u'stuff' )

        fixture.current_value = u'other'
        rendered = tester.render_html()
        vassert( rendered == u'other' )

    @test(WebFixture)
    def text_node_escapes_html(self, fixture):
        """The text of a TextNode is html-escaped."""

        widget = TextNode(fixture.view, u'<tag>')
        tester = WidgetTester(widget)
        
        rendered = tester.render_html()
        vassert( rendered == u'&lt;tag&gt;' )


    @test(WebFixture)
    def literal_html(self, fixture):
        """The LiteralHTML Widget just renders a chunk of HTML, but can answer queries about images in that HTML."""
        
        contents = u'<img src="_some_images/piet.pdf  "> <img src   = \' _some_images/koos was-^-hoêr.jpg\'>'
        literal_html = LiteralHTML(fixture.view, contents)
        tester = WidgetTester(literal_html)

        rendered_html = tester.render_html()
        vassert( rendered_html == contents )

        # Case: when the content is transformed
        def text_transformation(text):
            return text.replace(u'im', u'IM')
        literal_html = LiteralHTML(fixture.view, contents, transform=text_transformation)
        tester = WidgetTester(literal_html)

        rendered_html = tester.render_html()
        vassert( rendered_html == text_transformation(contents) )

    @test(WebFixture)
    def head(self, fixture):
        """Head corresponds with the head HTML element, can have a title and always has a special Slot used by the framework."""

        head = Head(fixture.view, u'a title')
        tester = WidgetTester(head)

        reahl_header_slot = head.children[1]
        vassert( isinstance(reahl_header_slot, Slot) )
        vassert( reahl_header_slot.name == u'reahl_header' )

        rendered_html = tester.render_html()
        vassert( rendered_html == u'<head><title>a title</title></head>' )

    @test(WebFixture)
    def body(self, fixture):
        """Body corresponds with the body HTML element, and always has a special Slot at its end used by the framework."""
        body = Body(fixture.view)
        tester = WidgetTester(body)

        body.add_child(P(fixture.view))
        
        reahl_footer_slot = body.children[1]
        vassert( isinstance(reahl_footer_slot, Slot) )
        vassert( reahl_footer_slot.name == u'reahl_footer' )

        rendered_html = tester.render_html()
        vassert( rendered_html == u'<body><p></p></body>' )







# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.




from nose.tools import istest
from reahl.tofu import Fixture, test, scenario
from reahl.tofu import vassert, expected

from reahl.component.modelinterface import Field, RequiredConstraint
from reahl.component.exceptions import ProgrammerError
from reahl.web.fw import ReahlWSGIApplication, UserInterface, UrlBoundView
from reahl.web.ui import TwoColumnPage, P, A
from reahl.webdev.tools import Browser, WidgetTester
from reahl.web_dev.fixtures import WebFixture

@istest
class ParameterisedViewErrors(object):
    @test(WebFixture)
    def missing_variable_in_regex(self, fixture):
        """If a variable is missing from the regex, an appropriate error is raised."""

        class ParameterisedView(UrlBoundView):
            def assemble(self, some_key=None):
                self.title = u'View for: %s' % some_key

        class UIWithParameterisedViews(UserInterface):
            def assemble(self):
                self.define_regex_view(u'/(?P<incorrect_name_for_key>.*)', u'/${key}', view_class=ParameterisedView, some_key=Field(required=True))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  UIWithParameterisedViews,  {}, name=u'test_ui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        def check_message(ex):
            return unicode(ex).startswith('The arguments contained in URL')
        with expected(ProgrammerError, test=check_message):
            browser.open('/a_ui/test1/')



class ParameterisedUserInterfaceErrors(WebFixture):
    def new_wsgi_app(self):
        fixture = self
        class RegexUserInterface(UserInterface):
            def assemble(self, ui_key=None):
                self.name = u'user_interface-%s' % ui_key

        class UIWithParameterisedUserInterfaces(UserInterface):
            def assemble(self):
                self.define_regex_user_interface(u'/(?P<xxx>[^/]*)', u'N/A', RegexUserInterface,
                                         {u'user_interface-slot': u'main'},
                                         ui_key=Field(required=True))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                self.define_user_interface(u'/a_ui',  UIWithParameterisedUserInterfaces,  {}, name=u'test_ui')

        return super(ParameterisedUserInterfaceErrors, self).new_wsgi_app(site_root=MainUI)
       

@istest
class ParameterisedErrorsTests(object):
    @test(ParameterisedUserInterfaceErrors)
    def missing_variable_in_regex(self, fixture):

        browser = Browser(fixture.wsgi_app)

        with expected(RequiredConstraint):
            browser.open('/a_ui/test1/')



# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from webob import Request
from nose.tools import istest
from reahl.tofu import Fixture, scenario, test
from reahl.tofu import vassert, expected
from reahl.stubble import stubclass, EmptyStub

from reahl.web.ui import *
from reahl.component.modelinterface import Field, BooleanField
from reahl.web.fw import Bookmark, Url
from reahl.webdev.tools import WidgetTester, XPath, Browser
from reahl.web_dev.widgets.widgetbasics import WidgetFixture
from reahl.web_dev.fixtures import WebFixture

_ = Translator(u'reahl-web')

@istest
class BasicReahlWidgets(object):
    @test(WebFixture)
    def priority_group(self, fixture):
        """A Priority group is a construct that ensures only one of its members has primary priority."""
        g = PriorityGroup()
        w1 = P(fixture.view)
        w2 = P(fixture.view)
        w3 = P(fixture.view)
        
        g.add_secondary(w1)
        g.add_primary(w2)
        g.add_secondary(w3)

        vassert( w1.attributes.v[u'class'] == u'reahl-priority-secondary' )
        vassert( w2.attributes.v[u'class'] == u'reahl-priority-primary' )
        vassert( w3.attributes.v[u'class'] == u'reahl-priority-secondary' )
        
        # You cannot have more than one primary
        w4 = P(fixture.view)
        with expected(AssertionError):
            g.add_primary(w4)

        # You cannot add something more than once
        with expected(AssertionError):
            g.add_secondary(w3)


    class Scenarios(WebFixture):
        @scenario
        def panel(self):
            self.widget = Panel(self.view)
            self.expected_html = u'<div></div>'

        @scenario
        def input_group1(self):
            self.widget = InputGroup(self.view)
            self.expected_html = u'<fieldset></fieldset>'

        @scenario
        def input_group2(self):
            self.widget = InputGroup(self.view, label_text=u'text')
            self.expected_html = u'<fieldset><label>text</label></fieldset>'

                                
    @test(Scenarios)
    def aliassed_widgets(self, fixture):
        """Some Widgets really also just render HTML, but have more abstract names."""
        tester = WidgetTester(fixture.widget)
        rendered_html = tester.render_html()
        vassert( rendered_html == fixture.expected_html )

    @test(WebFixture)
    def html5_page(self, fixture):
        """An HTML5Page is an empty HTML 5 document using the header and body widgets."""
        widget = HTML5Page(fixture.view, title=u'It: $current_title')
        widget.add_default_slot(u'slot1', P.factory())
        tester = WidgetTester(widget)
        
        rendered_html = tester.render_html()
        head = u'<head><title>It: %s</title></head>' % fixture.view.title
        vassert( rendered_html == u'<!DOCTYPE html><html>%s<body></body></html>' % head)
        
        vassert( widget.default_slot_definitions.keys() == [u'slot1'] )

    @test(WebFixture)
    def rendering_a_menu_item(self, fixture):
        """A MenuItem is a li with a in it."""
        
        description = 'The link'
        href_path = '/link'
        bookmark = Bookmark(u'/', href_path, description)

        menu_item = MenuItem.from_bookmark(fixture.view, bookmark)
        tester = WidgetTester(menu_item)

        with fixture.context:
            actual = tester.render_html()
            
        vassert( actual == u'<li><a href="%s">%s</a></li>' % (href_path, description) )
        
    @test(WebFixture)
    def menu(self, fixture):
        """Menus can be constructed from a list of A's or Bookmarks, or MenuItems can be added to them."""
        # Case: a normal menu from bookmarks
        item_specs = [Bookmark(u'/', u'/href1', u'description1'),
                      Bookmark(u'/', u'/href2', u'description2')]
        menu = Menu.from_bookmarks(fixture.view, item_specs)
        tester = WidgetTester(menu)

        with fixture.context:
            actual = tester.render_html()
        rendered_children = WidgetTester(menu.children).render_html()
        expected = u'''<ul class="reahl-menu">%s</ul>''' % rendered_children
        vassert( actual == expected )
        
        #case: using A's
        a_list = [A.from_bookmark(fixture.view, i) for i in item_specs]
        menu = Menu(fixture.view, a_list)
        tester = WidgetTester(menu)
        with fixture.context:
            actual = tester.render_html()
        vassert( actual == expected )
        
        # Case: adding already constructed menu item
        menu = Menu(fixture.view, [])
        menu.add_item(MenuItem(fixture.view, A.from_bookmark(fixture.view, item_specs[0])))
        menu.add_item(MenuItem(fixture.view, A.from_bookmark(fixture.view, item_specs[1])))
        vassert( menu.menu_items == menu.children )
        
        tester = WidgetTester(menu)
        with fixture.context:
            actual = tester.render_html()
        vassert( actual == expected )
        
        # Case: HMenu
        menu = HMenu.from_bookmarks(fixture.view, item_specs)
        vassert( menu.attributes.v[u'class'] == u'reahl-horizontal reahl-menu' )

        # Case: VMenu
        menu = VMenu.from_bookmarks(fixture.view, item_specs)
        vassert( menu.attributes.v[u'class'] == u'reahl-menu reahl-vertical' )

    @test(WebFixture)
    def menu_can_have_submenus(self, fixture):
        """Menus can have sub-menus too - when built-up using .add_item()."""
        # Case: a normal menu from bookmarks
        single_item_bookmark = Bookmark(u'/', u'/href3', u'description3')
        item_specs = [Bookmark(u'/', u'/href1', u'description1'),
                      Bookmark(u'/', u'/href2', u'description2')]

        sub_menu = Menu.from_bookmarks(fixture.view, item_specs)
        sub_menu_title = u'Subbie'
        menu = Menu(fixture.view, [])
        menu.add_item(MenuItem(fixture.view, A.from_bookmark(fixture.view, single_item_bookmark)))
        menu.add_item(SubMenu(fixture.view, sub_menu_title, sub_menu))

        expected_html = u'''<ul class="reahl-menu">'''\
                   u'''<li><a href="/href3">description3</a></li>'''\
                   u'''<li><a>Subbie</a>'''\
                   u'''<ul class="reahl-menu">'''\
                   u'''<li><a href="/href1">description1</a></li>'''\
                   u'''<li><a href="/href2">description2</a></li>'''\
                   u'''</ul>'''\
                   u'''</li>'''\
                   u'''</ul>'''
        tester = WidgetTester(menu)
        with fixture.context:
            actual = tester.render_html()
        vassert( actual == expected_html )


    class MenuItemScenarios(WebFixture):
        description = 'The link'
        href = Url('/link')

        @scenario
        def not_active(self):
            self.active_regex = None
            self.request.environ[u'PATH_INFO'] = u'/something/else'
            self.active = False

        @scenario
        def active_exact_path(self):
            self.active_regex = None
            self.request.environ[u'PATH_INFO'] = u'/link'
            self.active = True

        @scenario
        def active_partial_path(self):
            self.active_regex = None
            self.request.environ[u'PATH_INFO'] = u'/link/something/more'
            self.active = True

        @scenario
        def inactive_partial_path(self):
            self.active_regex = u'^/link$'
            self.request.environ[u'PATH_INFO'] = u'/link/something/more'
            self.active = False

    @test(MenuItemScenarios)
    def rendering_active_menu_items(self, fixture):    
        description = 'The link'
        href = Url('/link')
        
        menu_item = MenuItem(fixture.view, A(fixture.view, href, description=description), active_regex=fixture.active_regex)
        tester = WidgetTester(menu_item)

        actual = tester.render_html()
        class_str = u'' if not fixture.active else u' class="active"'
        expected_menu_item_html = u'<li%s><a href="/link">The link</a></li>' % (class_str)

        vassert( actual == expected_menu_item_html )


    @test(WebFixture)
    def language_menu(self, fixture):
        """A Menu can also be constructed to let a user choose to view the same page in 
           another of the supported languages."""
        
        class PanelWithMenu(Panel):
            def __init__(self, view):
                super(PanelWithMenu, self).__init__(view)
                self.add_child(HMenu.from_languages(view))
                self.add_child(P(view, text=_(u'This is an English sentence.')))

        wsgi_app = fixture.new_wsgi_app(child_factory=PanelWithMenu.factory())
        
        browser = Browser(wsgi_app)
        browser.open(u'/')
        
        vassert( browser.is_element_present(XPath.paragraph_containing(u'This is an English sentence.')) )

        browser.click(XPath.link_with_text(u'Afrikaans'))
        vassert( browser.is_element_present(XPath.paragraph_containing(u'Hierdie is \'n sin in Afrikaans.')) )

        browser.click(XPath.link_with_text(u'English (United Kingdom)'))
        vassert( browser.is_element_present(XPath.paragraph_containing(u'This is an English sentence.')) )


    class Scenarios(WebFixture):
        @scenario
        def yuidoc(self):
            self.widget = YuiDoc(self.view, u'docid', u'docclass')
            self.expected_html = u'<div id="docid" class="docclass"><div id="hd" class="yui-g"><header></header></div><div id="bd" role="main"><div id="yui-main"><div class="yui-b"></div></div><div class="yui-b"></div></div><div id="ft"><footer></footer></div></div>'
        @scenario
        def yuiblock(self):
            self.widget = YuiBlock(self.view)
            self.expected_html = u'<div class="yui-b"></div>'

        @scenario
        def yuigrid(self):
            self.widget = YuiGrid(self.view)
            self.expected_html = u'<div class="yui-g"></div>'

        @scenario
        def yuiunit1(self):
            self.widget = YuiUnit(self.view)
            self.expected_html = u'<div class="yui-u"></div>'

        @scenario
        def yuiunit1(self):
            self.widget = YuiUnit(self.view, first=True)
            self.expected_html = u'<div class="first yui-u"></div>'
            
    @test(Scenarios)
    def yui_widgets(self, fixture):
        """Different Yui widgets for rendering HTML that it compatible with the Yui CSS framework."""
        tester = WidgetTester(fixture.widget)
        rendered_html = tester.render_html()
        vassert( rendered_html == fixture.expected_html )
        
    @test(WebFixture)
    def twocolumn_page(self, fixture):
        """A simple Yui page with two columns, a header and a footer."""
        widget = TwoColumnPage(fixture.view, title=u'It: $current_title')
        widget.add_default_slot(u'slot1', P.factory())
        tester = WidgetTester(widget)
        
        rendered_html = tester.render_html()
        expected = u'<!DOCTYPE html><html><head><title>It: A view</title></head><body><div id="doc" class="yui-t2"><div id="hd" class="yui-g"><header></header></div><div id="bd" role="main"><div id="yui-main"><div class="yui-b"></div></div><div class="yui-b"></div></div><div id="ft"><footer></footer></div></div></body></html>'
        vassert( rendered_html == expected )
        
        vassert( widget.default_slot_definitions.keys() == [u'slot1'] )
        
        vassert( isinstance(widget.yui_page, YuiDoc) )
        vassert( widget.footer is widget.yui_page.footer )
        vassert( widget.header is widget.yui_page.header )
        vassert( widget.main is widget.yui_page.main_block )
        vassert( widget.secondary is widget.yui_page.secondary_block )




class TabbedPanelAjaxFixture(WebFixture):
    def new_PopulatedTabbedPanel(self):
        fixture = self
        class PopulatedTabbedPanel(TabbedPanel):
            def __init__(self, view):
                super(PopulatedTabbedPanel, self).__init__(view, u'tabbed')
                multi_tab = MultiTab(view, u'multitab name', u'multi-main', P.factory(text=u'main multi tab content'))
                tab1 = Tab(fixture.view, u'tab 1 name', u'multi1', P.factory(text=u'tab 1/1 content'))
                tab2 = Tab(fixture.view, u'tab 2 name', u'multi2', P.factory(text=u'tab 1/2 content'))
                multi_tab.add_tab(tab1)
                multi_tab.add_tab(tab2)
                self.add_tab(multi_tab)

                tab3 = Tab(fixture.view, u'tab 3 name', u'tab3', P.factory(text=u'tab 3 content'))
                self.add_tab(tab3)
        return PopulatedTabbedPanel

    def new_wsgi_app(self):
        return super(TabbedPanelAjaxFixture, self).new_wsgi_app(enable_js=True,
                                                   child_factory=self.PopulatedTabbedPanel.factory())
    def tab_is_active(self, tab_name):
        return self.driver_browser.execute_script('return window.jQuery("a:contains(\'%s\')").parent().hasClass("active")' % tab_name)

    def tab_contents_equals(self, expected_contents):
        return self.driver_browser.execute_script('return window.jQuery("div.reahl-tabbedpanel div").html() == "%s"' % expected_contents)


@istest
class TabbedPanelTests(object):
    @test(WebFixture)
    def basic_rendering(self, fixture):
        """A TabbedPanel is a Panel which contains an HMenu and a Panel."""
        fixture.request.query_string = u'tab=tab1'
        tabbed_panel = TabbedPanel(fixture.view, u'tabbed_name')
        tabbed_panel.add_tab(Tab(fixture.view, u'tab 1 name', u'tab1', P.factory(text=u'tab 1 content')))

        tester = WidgetTester(tabbed_panel)

        expected_html = u'''<div id="tabbed_name" class="reahl-tabbedpanel">'''\
                        u'''<ul class="reahl-horizontal reahl-menu">'''\
                        u'''<li class="active"><a href="?tab=tab1" class="reahl-ajaxlink">tab 1 name</a></li>'''\
                        u'''</ul>'''\
                        u'''<div><p>tab 1 content</p></div>'''\
                        u'''</div>'''
        actual = tester.render_html()        
        vassert( actual == expected_html )

    @test(WebFixture)
    def tabs_with_sub_options(self, fixture):
        """A TabbedPanel can have Tabs that are each composed of multiple sub-options."""
        fixture.request.query_string = u'tab=mult2'
        tabbed_panel = TabbedPanel(fixture.view, u'tabbed_name')
        multi_tab = MultiTab(fixture.view, u'tab 1 name', u'multitab-main', P.factory(text=u'multi tab main content'))
        multi_tab.add_tab(Tab(fixture.view, u'multi tab 1', u'mult1', P.factory(text=u'tab 1/1 content')))
        multi_tab.add_tab(Tab(fixture.view, u'multi tab 2', u'mult2', P.factory(text=u'tab 1/2 content')))
        tabbed_panel.add_tab(multi_tab)

        tester = WidgetTester(tabbed_panel)
        
        expected_html = u'''<div id="tabbed_name" class="reahl-tabbedpanel">'''\
                        u'''<ul class="reahl-horizontal reahl-menu">'''\
                        u'''<li class="active"><a href="?tab=multitab-main" class="reahl-ajaxlink">tab 1 name</a>&nbsp;'''\
                        u'''<a class="dropdown-handle">▼</a>'''\
                        u'''<ul class="reahl-menu reahl-vertical">'''\
                        u'''<li><a href="?tab=mult1" class="reahl-ajaxlink">multi tab 1</a></li>'''\
                        u'''<li class="active"><a href="?tab=mult2" class="reahl-ajaxlink">multi tab 2</a></li>'''\
                        u'''</ul>'''\
                        u'''</li>'''\
                        u'''</ul>'''\
                        u'''<div><p>tab 1/2 content</p></div>'''\
                        u'''</div>'''
        actual = tester.render_html()        
        vassert( actual == expected_html )

    class DefaultTabScenarios(WebFixture):
        @scenario
        def specified_on_query_string(self):
            self.request.query_string = u'tab=tab2'
            self.expected_contents = u'<p>tab 2 content</p>'
            self.tab1_active = False
            self.tab2_active = True
            
        @scenario
        def defaulted(self):
            self.request.query_string = u''
            self.expected_contents = u'<p>tab 1 content</p>'
            self.tab1_active = True
            self.tab2_active = False
            
    @test(DefaultTabScenarios)
    def default_active_tab(self, fixture):
        """The first tab is active by default (if the active tab is not indicated in the query_string)."""
        tab1 = Tab(fixture.view, u'tab 1 name', u'tab1', P.factory(text=u'tab 1 content'))
        tab2 = Tab(fixture.view, u'tab 2 name', u'tab2', P.factory(text=u'tab 2 content'))

        tabbed_panel = TabbedPanel(fixture.view, u'tabbed_name')
        tabbed_panel.add_tab(tab1)
        tabbed_panel.add_tab(tab2)

        vassert( tab1.is_active == fixture.tab1_active )
        vassert( tab2.is_active == fixture.tab2_active )

        tester = WidgetTester(tabbed_panel)
        panel_contents = tester.get_html_for('//div/div/*')
        vassert( panel_contents == fixture.expected_contents )

    class DefaultMultiTabScenarios(WebFixture):
        @scenario
        def specified_on_query_string(self):
            self.request.query_string = u'tab=multi2'
            self.expected_contents = u'<p>tab 1/2 content</p>'
            self.multi_tab_active = True
            self.tab1_active = False
            self.tab2_active = True
            self.tab3_active = False
            
        @scenario
        def defaulted(self):
            self.request.query_string = u''
            self.expected_contents = u'<p>multi tab main content</p>'
            self.multi_tab_active = True
            self.tab1_active = False
            self.tab2_active = False
            self.tab3_active = False

        @scenario
        def specified_to_other_tab(self):
            self.request.query_string = u'tab=tab3'
            self.expected_contents = u'<p>tab 3 content</p>'
            self.multi_tab_active = False
            self.tab1_active = False
            self.tab2_active = False
            self.tab3_active = True

    @test(DefaultMultiTabScenarios)
    def default_active_multi_tab(self, fixture):
        """The first item of the first tab is active by default (if the active tab is not indicated in the query_string)."""
        tabbed_panel = TabbedPanel(fixture.view, u'tabbed_panel')
        multi_tab = MultiTab(fixture.view, u'tab 1 name', u'multi-tab', P.factory(text=u'multi tab main content'))
        tab1 = Tab(fixture.view, u'tab 1 name', u'multi1', P.factory(text=u'tab 1/1 content'))
        tab2 = Tab(fixture.view, u'tab 2 name', u'multi2', P.factory(text=u'tab 1/2 content'))
        multi_tab.add_tab(tab1)
        multi_tab.add_tab(tab2)
        tabbed_panel.add_tab(multi_tab)

        tab3 = Tab(fixture.view, u'tab 3 name', u'tab3', P.factory(text=u'tab 3 content'))
        tabbed_panel.add_tab(tab3)

        vassert( multi_tab.is_active == fixture.multi_tab_active )
        vassert( tab1.is_active == fixture.tab1_active )
        vassert( tab2.is_active == fixture.tab2_active )
        vassert( tab3.is_active == fixture.tab3_active )

        tester = WidgetTester(tabbed_panel)
        panel_contents = tester.get_html_for('//div/div/*')
        vassert( panel_contents == fixture.expected_contents )

    @test(TabbedPanelAjaxFixture)
    def switching_panels(self, fixture):
        """The contents registered with the TabbedPanel's active tab are displayed."""
        fixture.reahl_server.set_app(fixture.wsgi_app)
        fixture.driver_browser.open(u'/')

        # by default, the first tab is active and its contents are displayed
        vassert( fixture.driver_browser.wait_for(fixture.tab_is_active, u'multitab name') )
        vassert( fixture.driver_browser.wait_for(fixture.tab_contents_equals, u'<p>main multi tab content</p>') )

        fixture.driver_browser.click(XPath.link_with_text(u'tab 3 name'))
        # tab3 is active and its contents are shown
        vassert( fixture.driver_browser.wait_for(fixture.tab_is_active, u'tab 3 name') )
        vassert( fixture.driver_browser.wait_for(fixture.tab_contents_equals, u'<p>tab 3 content</p>') )

        fixture.driver_browser.mouse_over(XPath.link_with_text(u'multitab name'))
        fixture.driver_browser.click(XPath.link_with_text(u'tab 2 name'))
        # tab2 is active and its contents are shown
        vassert( fixture.driver_browser.wait_for(fixture.tab_is_active, u'tab 2 name') )
        vassert( fixture.driver_browser.wait_for(fixture.tab_contents_equals, u'<p>tab 1/2 content</p>') )

        fixture.driver_browser.click(XPath.link_with_text(u'▼'))
        # clicking on the down arrow does nothing
        vassert( fixture.driver_browser.wait_for(fixture.tab_is_active, u'tab 2 name') )

        fixture.driver_browser.click(XPath.link_with_text(u'multitab name'))
        # the main multitab is active and its contents are shown
        vassert( fixture.driver_browser.wait_for(fixture.tab_is_active, u'multitab name') )
        vassert( fixture.driver_browser.wait_for(fixture.tab_contents_equals, u'<p>main multi tab content</p>') )



class SlidingPanelFixture(WebFixture):
    def new_PopulatedSlidingPanel(self):
        fixture = self
        class PopulatedSlidingPanel(SlidingPanel):
            def __init__(self, view):
                super(PopulatedSlidingPanel, self).__init__(view, u'slide')
                panel0 = Panel(view)
                panel0.add_child(P(view, text=u'Contents for panel 0'))
                self.add_panel(panel0)

                panel1 = Panel(view)
                panel1.add_child(P(view, text=u'Contents for panel 1'))
                self.add_panel(panel1)

        return PopulatedSlidingPanel

    def new_wsgi_app(self):
        return super(SlidingPanelFixture, self).new_wsgi_app(enable_js=True,
                                                           child_factory=self.PopulatedSlidingPanel.factory())
    def panel_is_visible(self, index):
        return self.driver_browser.wait_for_element_visible(XPath.paragraph_containing(u'Contents for panel %s' % index))

    def panel_is_not_visible(self, index):
        return self.driver_browser.wait_for_element_not_visible(XPath.paragraph_containing(u'Contents for panel %s' % index))



@test(SlidingPanelFixture)
def opens_on_selected_index(fixture):
    fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = fixture.driver_browser
    
    browser.open(u'/?index=1')
    vassert( fixture.panel_is_visible(1) )
    

@test(SlidingPanelFixture)
def slide_right(fixture):
    fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = fixture.driver_browser

    browser.open(u'/')
    vassert( fixture.panel_is_visible(0) )

    browser.click(XPath.link_with_text(u'>'))
    vassert( fixture.panel_is_not_visible(0) )
    vassert( fixture.panel_is_visible(1) )

    #wrapping back to start
    browser.click(XPath.link_with_text(u'>'))
    vassert( fixture.panel_is_not_visible(1) )
    vassert( fixture.panel_is_visible(0) )

@test(SlidingPanelFixture)
def slide_left(fixture):
    fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = fixture.driver_browser

    browser.open(u'/')
    vassert( fixture.panel_is_visible(0) )

    #wrapping
    browser.click(XPath.link_with_text(u'<'))
    vassert( fixture.panel_is_not_visible(0) )
    vassert( fixture.panel_is_visible(1) )

    browser.click(XPath.link_with_text(u'<'))
    vassert( fixture.panel_is_not_visible(1) )
    vassert( fixture.panel_is_visible(0) )

class PopupAFixture(WebFixture):
    # (note that this xpath ensures that the p is the ONLY content of the dialog)
    poppedup_contents = u"//div[@class='reahl-dialogcontent' and count(*)=1]/p[@id='contents']"
    

@istest
class PopupATests(object):
    @test(PopupAFixture)
    def default_behaviour(self, fixture):
        """If you click on the A, a popupwindow opens with its contents the specified
           element on the target page."""

        class PopupTestPanel(Panel):
            def __init__(self, view):
                super(PopupTestPanel, self).__init__(view)
                self.add_child(PopupA(view, view.as_bookmark(), u'#contents'))
                popup_contents = self.add_child(P(view, text=u'this is the content of the popup'))
                popup_contents.set_id(u'contents')

        wsgi_app = fixture.new_wsgi_app(child_factory=PopupTestPanel.factory(), enable_js=True)
        fixture.reahl_server.set_app(wsgi_app)
        fixture.driver_browser.open('/')
        
        # The A is rendered correctly
        fixture.driver_browser.is_element_present(u"//a[@title='Home page' and text()='Home page' and @href='/']")

        # subsequent behaviour
        fixture.driver_browser.click(XPath.link_with_text(u'Home page'))
        fixture.driver_browser.wait_for_element_visible(fixture.poppedup_contents)
        
        fixture.driver_browser.click(XPath.button_labelled(u'Close'))
        fixture.driver_browser.wait_for_element_not_visible(fixture.poppedup_contents)

    @test(PopupAFixture)
    def customising_dialog_buttons(self, fixture):
        """The buttons of the dialog can be customised."""
        
        class PopupTestPanel(Panel):
            def __init__(self, view):
                super(PopupTestPanel, self).__init__(view)
                popup_a = self.add_child(PopupA(view, view.as_bookmark(), u'#contents'))
                popup_a.add_button(DialogButton(u'Butt1'))
                popup_a.add_button(DialogButton(u'Butt2'))
                popup_contents = self.add_child(P(view, text=u'this is the content of the popup'))
                popup_contents.set_id(u'contents')

        wsgi_app = fixture.new_wsgi_app(child_factory=PopupTestPanel.factory(), enable_js=True)
        fixture.reahl_server.set_app(wsgi_app)

        button1_xpath = XPath.button_labelled(u'Butt1')
        button2_xpath = XPath.button_labelled(u'Butt2')
        fixture.driver_browser.open('/')

        fixture.driver_browser.click(XPath.link_with_text(u'Home page'))
        fixture.driver_browser.wait_for_element_visible(fixture.poppedup_contents)

        vassert( fixture.driver_browser.is_element_present(button1_xpath) )
        vassert( fixture.driver_browser.is_element_present(button2_xpath) )

    @test(PopupAFixture)
    def workings_of_check_checkbox_button(self, fixture):
        """A CheckCheckBoxButton checks the checkbox on the original page when clicked."""
        
        class PopupTestPanel(Panel):
            @exposed
            def fields(self, fields):
                fields.field = BooleanField()
                
            def __init__(self, view):
                super(PopupTestPanel, self).__init__(view)
                popup_a = self.add_child(PopupA(view, view.as_bookmark(), u'#contents'))
                popup_contents = self.add_child(P(view, text=u'this is the content of the popup'))
                popup_contents.set_id(u'contents')
                form = self.add_child(Form(view, u'aform'))
                checkbox = form.add_child(CheckboxInput(form, self.fields.field))

                popup_a.add_button(CheckCheckboxButton(u'Checkit', checkbox))

        wsgi_app = fixture.new_wsgi_app(child_factory=PopupTestPanel.factory(), enable_js=True)
        fixture.reahl_server.set_app(wsgi_app)

        button_xpath = XPath.button_labelled(u'Checkit')
        fixture.driver_browser.open('/')

        fixture.driver_browser.click(XPath.link_with_text(u'Home page'))
        fixture.driver_browser.wait_for_element_visible(fixture.poppedup_contents)

        fixture.driver_browser.click(button_xpath)
        fixture.driver_browser.wait_for_element_not_visible(fixture.poppedup_contents)

        vassert( fixture.driver_browser.is_checked(u"//input[@type='checkbox']") )







# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



from webob import Response
from webob.exc import HTTPMethodNotAllowed

from reahl.stubble import stubclass, exempt
from nose.tools import istest
from reahl.tofu import test, Fixture, scenario
from reahl.tofu import vassert

from reahl.component.modelinterface import Field
from reahl.webdev.tools import Browser
from reahl.web.fw import Resource, SubResource, UserInterface, Widget
from reahl.web_dev.fixtures import  WebFixture


@istest
class SubResourcesTests(object):
    @test(WebFixture)
    def resources(self, fixture):
        """A Resource has the responsibility to handle an HTTP Request. A Resource indicates that it can
           handle a particular HTTP method by having a method named for it. Such method should return
           a Response."""

        @stubclass(Resource)
        class ResourceStub(Resource):
            called = None
            @exempt
            def handle_something(self, request):
                self.called = True
                return u'something'
            @exempt
            def handle_anotherthing(self, request):
                pass
                 
        resource = ResourceStub()
        
        # Case where the HTTP method is not supported
        fixture.request.method = 'koos'
        response = resource.handle_request(fixture.request)
        vassert( isinstance(response, HTTPMethodNotAllowed) )
        vassert( not resource.called )
        vassert( response.headers['allow'] == 'anotherthing, something' )
        
        # Case where the HTTP method is supported
        fixture.request.method = 'SOMEthing'
        response = resource.handle_request(fixture.request)
        vassert( resource.called )
        vassert( response == u'something' )

    @test(WebFixture)
    def simple_sub_resources(self, fixture):
        """During their construction, Widgets can add SubResources to their View.  The SubResource
           will then be available via a special URL underneath the URL of the Widget's View."""
    
        @stubclass(SubResource)
        class ASimpleSubResource(SubResource):
            sub_regex = u'simple_resource'
            sub_path_template = u'simple_resource'
            @exempt
            def handle_get(self, request):
                return Response()

        @stubclass(Widget)
        class WidgetWithSubResource(Widget):
            def __init__(self, view):
                super(WidgetWithSubResource, self).__init__(view)
                view.add_resource(ASimpleSubResource(u'uniquename'))

        wsgi_app = fixture.new_wsgi_app(view_slots={u'main': WidgetWithSubResource.factory()})
        browser = Browser(wsgi_app)
        browser.open('/_uniquename_simple_resource')
        
    @test(WebFixture)
    def dynamic_sub_resources(self, fixture):
        """Sometimes it is undesirable to instantiate a SubResource just for the purpose of adding it 
           to a View. In such cases, A Factory can be added that will construct the SubResource
           on demand, based on a regex which may contain parameters."""
        @stubclass(SubResource)
        class ParameterisedSubResource(SubResource):
            sub_regex = u'dynamic_(?P<param>[^/]+)'
            sub_path_template = u'dynamic_%(param)s'

            def __init__(self, unique_name, param):
                super(ParameterisedSubResource, self).__init__(unique_name)
                self.param = param
                
            @exempt
            def handle_get(self, request):
                return Response(body=str(self.param))

            @exempt
            @classmethod
            def create_resource(cls, unique_name, param=None):
                return cls(unique_name, param)

        @stubclass(Widget)
        class WidgetWithSubResource(Widget):
            def __init__(self, view):
                super(WidgetWithSubResource, self).__init__(view)
                view.add_resource_factory(ParameterisedSubResource.factory(u'uniquename', {u'param': Field(required=True)}))


        wsgi_app = fixture.new_wsgi_app(view_slots={u'main': WidgetWithSubResource.factory()})
        browser = Browser(wsgi_app)
        browser.open('/_uniquename_dynamic_one')
        vassert( browser.raw_html == u'one' )
        browser.open('/_uniquename_dynamic_two')
        vassert( browser.raw_html == u'two' )

    @test(WebFixture)
    def dynamic_sub_resources_factory_args(self, fixture):
        """Such dynamic SubResources can also be created with arguments specified to its Factory (instead of 
           only from the path)."""
        @stubclass(SubResource)
        class ParameterisedSubResource(SubResource):
            sub_regex = u'dynamic_(?P<path_param>[^/]+)'
            sub_path_template = u'dynamic_%(path_param)s'

            def __init__(self, unique_name, factory_arg, factory_kwarg, path_param):
                super(ParameterisedSubResource, self).__init__(unique_name)
                self.path_param = path_param
                self.factory_arg = factory_arg
                self.factory_kwarg = factory_kwarg
                
            @exempt
            def handle_get(self, request):
                args = u'%s|%s|%s' % (self.factory_arg, self.factory_kwarg, self.path_param)
                return Response(body=str(args))

            @exempt
            @classmethod
            def create_resource(cls, unique_name, factory_arg, factory_kwarg=None, path_param=None):
                return cls(unique_name, factory_arg, factory_kwarg, path_param)

        @stubclass(Widget)
        class WidgetWithSubResource(Widget):
            def __init__(self, view):
                super(WidgetWithSubResource, self).__init__(view)
                factory = ParameterisedSubResource.factory(u'uniquename', {u'path_param': Field(required=True)}, u'arg to factory', factory_kwarg=u'kwarg to factory')
                view.add_resource_factory(factory)

        wsgi_app = fixture.new_wsgi_app(view_slots={u'main': WidgetWithSubResource.factory()})
        browser = Browser(wsgi_app)
        browser.open('/__uniquename_dynamic_one')
        vassert( browser.raw_html == u'arg to factory|kwarg to factory|one' )


    @test(WebFixture)
    def disambiguating_between_factories(self, fixture):
        """Sometimes, Widgets may need to add more than one SubResource (or Factories) of the same type.
           Since these are of the same type, they will both match the same urls based on their sub_regex.
           For this reason, one can use the unique_name passed to a SubResource or its Factory
           to disambiguate."""
        @stubclass(SubResource)
        class ParameterisedSubResource(SubResource):
            sub_regex = u'path'
            sub_path_template = u'path'

            @exempt
            def handle_get(self, request):
                # We send back the unique_name of the factory so we can test which factory handled the request
                return Response(body=str(self.unique_name))

            @exempt
            @classmethod
            def create_resource(cls, unique_name):
                return cls(unique_name)

        @stubclass(Widget)
        class WidgetWithAmbiguousSubResources(Widget):
            def __init__(self, view):
                super(WidgetWithAmbiguousSubResources, self).__init__(view)
                factory1 = ParameterisedSubResource.factory(u'factory1', {})
                view.add_resource_factory(factory1)
                factory2 = ParameterisedSubResource.factory(u'factory2', {})
                view.add_resource_factory(factory2)

        wsgi_app = fixture.new_wsgi_app(view_slots={u'main': WidgetWithAmbiguousSubResources.factory()})
        browser = Browser(wsgi_app)
        browser.open('/__factory1_path')
        vassert( browser.raw_html == u'factory1' )
        browser.open('/__factory2_path')
        vassert( browser.raw_html == u'factory2' )


    class UrlScenarios(WebFixture):
        @scenario
        def ending_in_slash(self):
            self.view_path = '/a/'
            self.expected_path = '/a/__uniquename_sub_path'
        @scenario
        def not_ending_in_slash(self):
            self.view_path = '/a'
            self.expected_path = '/a/_uniquename_sub_path'
        @scenario
        def only_slash(self):
            self.view_path = '/'
            self.expected_path = '/__uniquename_sub_path'
        
    @test(UrlScenarios)
    def computation_of_url(self, fixture):
        """The URL of a SubResource can be different, depending on the scenario.  This is done so that 
           A SubResource can unambiguously determine the URL of its parent View."""
        @stubclass(SubResource)
        class ASimpleSubResource(SubResource):
            sub_regex = u'sub_path'
            sub_path_template = u'sub_path'

        sub_resource = ASimpleSubResource(u'uniquename')

        # Calculating the sub_resource's URL from the context of the View
        fixture.request.environ['PATH_INFO'] = fixture.view_path
        actual = sub_resource.get_url()
        vassert( actual.path == fixture.expected_path )

        # Calculating the URL of the parent View from the context of the SubResource
        fixture.request.environ['PATH_INFO'] = fixture.expected_path
        actual = SubResource.get_parent_url()
        vassert( actual.path == fixture.view_path )

        # Calculating the sub_resource's URL from the context of the SubResource
        fixture.request.environ['PATH_INFO'] = fixture.expected_path
        actual = sub_resource.get_url()
        vassert( actual.path == fixture.expected_path )




# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import json

from webob import Request

from nose.tools import istest
from reahl.tofu import Fixture, test, scenario, NoException, vassert, expected
from reahl.stubble import CallMonitor, EmptyStub

from reahl.web.ui import Form, NestedForm, Button, Input, TextInput, LabelledBlockInput, TwoColumnPage
from reahl.web.fw import WebExecutionContext, ValidationException, Url, UserInterface
from reahl.component.exceptions import DomainException, ProgrammerError, IsInstance
from reahl.component.modelinterface import IntegerField, BooleanField, EmailField, DateField, \
                                    exposed, Field, Event, Action
from reahl.component.context import ExecutionContext
from reahl.webelixirimpl import PersistedException, UserInput

from reahl.web_dev.fixtures import WebBasicsMixin
from reahl.webdev.tools import WidgetTester, XPath, Browser

class FormFixture(Fixture, WebBasicsMixin):
    def new_error_xpath(self):
        return u'//form[contains(@class, "reahl-form")]/label[contains(@class, "error")]'


@istest
class FormTests(object):
    @test(FormFixture)
    def basic_event_linkup(self, fixture):
        """When a user clicks on a Button, the Event to which the Button is linked is triggered on 
           the server, and its corresponding action is executed. After this, the 
           browser is transitioned to the target view as specified by the EventHandler.
        """
        class ModelObject(object):
            handled_event = False
            def handle_event(self):
                self.handled_event = True

            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me', action=Action(self.handle_event))

        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name, other_view):
                super(MyForm, self).__init__(view, name)
                self.define_event_handler(model_object.events.an_event, target=other_view)
                self.add_child(Button(self, model_object.events.an_event))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/', title=u'Home page')
                other_view = self.define_view(u'/page2', title=u'Page 2')
                home.set_slot(u'main', MyForm.factory(u'myform', other_view))

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        fixture.reahl_server.set_app(wsgi_app)
        fixture.driver_browser.open('/')

        # clicking on the button triggers the action of the event handler
        vassert( not model_object.handled_event )
        fixture.driver_browser.click(u"//input[@value='click me']")
        vassert( model_object.handled_event )

        # browser has been transitioned to target view
        vassert( fixture.driver_browser.current_url.path == u'/page2' )

    @test(FormFixture)
    def arguments_to_actions(self, fixture):
        """If a Button is created for an Event with_arguments, those arguments are passed to the backend 
           when the Button is clicked."""

        # how you link everything up in code
        class ModelObject(object):
            def handle_event(self, *args):
                self.args = args

            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me',
                                        action=Action(self.handle_event, [u'one_argument', u'another_argument']),
                                        one_argument=IntegerField(),
                                        another_argument=Field())

        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                self.add_child(Button(self, model_object.events.an_event.with_arguments(one_argument=1, another_argument=u'another')))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/', title=u'Home page')
                other_view = self.define_view(u'/page2', title=u'Page 2')
                home.set_slot(u'main', MyForm.factory(u'myform'))
                self.define_transition(model_object.events.an_event, home, other_view)

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        fixture.reahl_server.set_app(wsgi_app)
        fixture.driver_browser.open('/')

        # when the Action is executed, the correct arguments are passed
        fixture.driver_browser.click(u"//input[@value='click me']")
        vassert( model_object.args == (1, u'another') )


    @test(FormFixture)
    def validation_of_event_arguments(self, fixture):
        """Buttons cannot be created for Events with invalid default arguments."""

        class ModelObject(object):
            @exposed
            def events(self, events):
                events.an_event = Event(label=u'Click me', argument=Field(required=True))

        model_object = ModelObject()
        
        form = Form(fixture.view, u'test')
        form.define_event_handler(model_object.events.an_event)
        
        with expected(ProgrammerError):
            Button(form, model_object.events.an_event)
            
        with expected(NoException):
            Button(form, model_object.events.an_event.with_arguments(argument=u'something'))


      
    @test(FormFixture)
    def basic_field_linkup(self, fixture):
        """When a form is rendered, the Inputs on it display either the default value as specified
           by the Field with which the Input is associated, or (if set) the value of the attribute
           that Field is bound to.
           
           When a user clicks on a Button, the Input value is set via the Field on the
           model objects to which the Field is bound. This happens before the associated action
           is executed.
        """

        class ModelObject(object):
            def handle_event(self):
                pass

            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me', action=Action(self.handle_event))

            @exposed
            def fields(self, fields):
                fields.field_name = IntegerField(default=3)
                
        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                self.define_event_handler(model_object.events.an_event)
                self.add_child(Button(self, model_object.events.an_event))
                self.add_child(TextInput(self, model_object.fields.field_name))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/', title=u'Home page')
                home.set_slot(u'main', MyForm.factory(u'myform'))


        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        fixture.reahl_server.set_app(wsgi_app)
        fixture.driver_browser.open('/')

        # the initial value rendered in the input
        vassert( not hasattr(model_object, u'field_name') )
        initial_value = fixture.driver_browser.get_value(u"//input[@type='text']")
        vassert( initial_value == u'3' )

        # the user supplied value is marshalled and set on the model object when an event happens
        fixture.driver_browser.type(u"//input[@type='text']", u'5')
        fixture.driver_browser.click(u"//input[@value='click me']")
        vassert( model_object.field_name == 5 )

    @test(FormFixture)
    def wrong_arguments_to_define_event_handler(self, fixture):
        """Passing anything other than an Event to define_event_handler is an error."""

        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                self.define_event_handler(EmptyStub())

        wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory(u'form'))
        browser = Browser(wsgi_app)
        
        with expected(IsInstance):
            browser.open('/')

    @test(FormFixture)
    def define_event_handler_not_called(self, fixture):
        """."""
        class ModelObject(object):
            @exposed
            def events(self, events):
                events.an_event = Event()

        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                self.add_child(Button(self, model_object.events.an_event))

        wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory(u'form'))
        browser = Browser(wsgi_app)

        def check_exc(exc):
            vassert( str(exc) == u'no Event/Transition available for name an_event' )

        with expected(ProgrammerError, test=check_exc):
            browser.open('/')


    @test(FormFixture)
    def exception_handling(self, fixture):
        """When a DomainException happens during the handling of an Event:

           The database is rolled back.
           The browser is redirected to GET the original view again (not the target).
           The screen still displays the values the user initially typed, not those on the ModelObject.
        """

        class ModelObject(object):
            def handle_event(self):
                self.field_name = 1
                raise DomainException()
            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me', action=Action(self.handle_event))
            @exposed
            def fields(self, fields):
                fields.field_name = IntegerField(default=3)

        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name, other_view):
                super(MyForm, self).__init__(view, name)
                self.define_event_handler(model_object.events.an_event, target=other_view)
                self.add_child(Button(self, model_object.events.an_event))
                self.add_child(TextInput(self, model_object.fields.field_name))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/', title=u'Home page')
                other_view = self.define_view(u'/page2', title=u'Page 2')
                home.set_slot(u'main', MyForm.factory(u'myform', other_view))

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        fixture.reahl_server.set_app(wsgi_app)

        fixture.driver_browser.open('/')

        vassert( not hasattr(model_object, u'field_name') )
        fixture.driver_browser.type(u"//input[@type='text']", u'5')

        # any database stuff that happened when the form was submitted was rolled back
        with CallMonitor(fixture.system_control.orm_control.rollback) as monitor:
            fixture.driver_browser.click(u"//input[@value='click me']")
        vassert( monitor.times_called == 1 )

        # the value input by the user is still displayed on the form, NOT the actual value on the model object        
        vassert( model_object.field_name == 1 )
        retained_value = fixture.driver_browser.get_value(u"//input[@type='text']")
        vassert( retained_value == u'5' )

        # the browser is still on the page with the form which triggered the exception
        vassert( fixture.driver_browser.current_url.path == u'/' )

    @test(FormFixture)
    def rendering_of_form(self, fixture):
        """A Form is always set up to POST to its EventChannel url.  The current page's query string is
           propagated with the POST url if any.  The Form has an id and class to help style it etc."""

        form = Form(fixture.view, u'test_channel')
        tester = WidgetTester(form)
        
        fixture.context.set_request(Request.blank(u'/a/b?x=y', charset='utf8'))
        actual = tester.render_html()

        expected = u'<form id="test_channel" action="/a/b/_test_channel_method?x=y" data-formatter="/__test_channel_format_method" method="POST" class="reahl-form"></form>'
        vassert( actual == expected )
        
        # Case: without querystring
        fixture.context.set_request(Request.blank(u'/a/b', charset='utf8'))
        actual = tester.render_html_tree()

        action = actual.xpath(u'//form')[0].attrib[u'action']
        vassert( action == u'/a/b/_test_channel_method' )
        
    @test(FormFixture)
    def duplicate_forms(self, fixture):
        """It is an error to add more than one form with the same unique_name to a page."""

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/', title=u'Home page')
                home.set_slot(u'main', Form.factory(u'myform'))
                home.set_slot(u'secondary', Form.factory(u'myform'))

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        def check_exc(ex):
            vassert( str(ex).startswith(u'More than one form was added using the same unique_name') )
        
        with expected(ProgrammerError, test=check_exc):
            browser.open('/')


    @test(FormFixture)
    def nested_forms(self, fixture):
        """HTML disallows nesting of forms. A NestedForm can be used to simulate 
           a form which is visually part of another Form. A NestedForm provides a
           form attribute which should be used when creating an Inputs for it.
        """

        class NestedModelObject(object):
            handled_event = False
            def handle_event(self):
                self.handled_event = True
            @exposed
            def events(self, events):
                events.nested_event = Event(label=u'click nested', action=Action(self.handle_event))
            @exposed
            def fields(self, fields):
                fields.nested_field = Field(label=u'input nested')

        nested_model_object = NestedModelObject()
        class MyNestedForm(NestedForm):
            def __init__(self, view, name):
                super(MyNestedForm, self).__init__(view, name)
                self.define_event_handler(nested_model_object.events.nested_event)
                self.add_child(Button(self.form, nested_model_object.events.nested_event))
                self.add_child(LabelledBlockInput(TextInput(self.form, nested_model_object.fields.nested_field)))

        class OuterModelObject(object):
            handled_event = False
            def handle_event(self):
                self.handled_event = True
            @exposed
            def events(self, events):
                events.outer_event = Event(label=u'click outer', action=Action(self.handle_event))
        outer_model_object = OuterModelObject()
        class OuterForm(Form):
            def __init__(self, view, name):
                super(OuterForm, self).__init__(view, name)
                self.add_child(MyNestedForm(view, u'my_nested_form'))
                self.define_event_handler(outer_model_object.events.outer_event)
                self.add_child(Button(self, outer_model_object.events.outer_event))

        wsgi_app = fixture.new_wsgi_app(child_factory=OuterForm.factory(u'outer_form'))
        fixture.reahl_server.set_app(wsgi_app)
        browser = fixture.driver_browser
        
        browser.open('/')
        browser.type(XPath.input_labelled(u'input nested'), u'some nested input')
        
        browser.click(XPath.button_labelled(u'click nested'))
        
        vassert( nested_model_object.handled_event )
        vassert( not outer_model_object.handled_event )
        
        vassert( nested_model_object.nested_field == u'some nested input' )
        

    @test(FormFixture)
    def form_input_validation(self, fixture):
        """Validation of input happens in JS on the client, but also on the server if JS is bypassed."""
        class ModelObject(object):
            def handle_event(self):
                pass
            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me', action=Action(self.handle_event))
            @exposed
            def fields(self, fields):
                fields.field_name = EmailField()

        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name, other_view):
                super(MyForm, self).__init__(view, name)
                self.define_event_handler(model_object.events.an_event, target=other_view)
                self.add_child(Button(self, model_object.events.an_event))
                self.add_child(TextInput(self, model_object.fields.field_name))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/', title=u'Home page')
                other_view = self.define_view(u'/page2', title=u'Page 2')
                home.set_slot(u'main', MyForm.factory(u'myform', other_view))

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI, enable_js=True)

        # Case: form validation fails in JS on the client
        #  - Form submission is blocked
        #  - Error message is displayed
        fixture.reahl_server.set_app(wsgi_app)
        fixture.driver_browser.open(u'/')
        fixture.driver_browser.wait_for_element_not_visible(fixture.error_xpath)
        fixture.driver_browser.type(u'//input[@type="text"]', u'not@notvalid')
        fixture.driver_browser.wait_for_element_visible(fixture.error_xpath)

        with fixture.driver_browser.no_page_load_expected():
            fixture.driver_browser.click(u"//input[@value='click me']")

        error_text = fixture.driver_browser.get_text(fixture.error_xpath)
        vassert( error_text == u'field_name should be a valid email address' )

        # Case: form validation fails on the server (assuming no JS on the client to block submission)
        #  - ValidationException is raised (which is dealt with as any DomainException)
        browser = Browser(wsgi_app)
        browser.open('/')
        browser.type('//input[@type="text"]', u'not@notvalid')

        browser.click('//input[@type="submit"]')
        vassert( not hasattr(model_object, 'field_name') )        
        label = browser.get_html_for('//label')
        vassert( label == u'<label for="field_name" class="error">field_name should be a valid email address</label>' )

        vassert( UserInput.query.filter_by(key=u'field_name').count() == 1 ) # The invalid input was persisted
        exception = PersistedException.query.one().exception
        vassert( isinstance(exception, ValidationException) ) # Is was persisted
        vassert( not exception.commit )

        # Case: form validation passes (no-js)
        #  - no ValidationException
        #  - all input is translated to python and set as values on the model objects
        #  - any saved input on the form is cleared
        browser.type(u'//input[@type="text"]', u'valid@home.org')
        browser.click(u"//input[@value='click me']")
        vassert( model_object.field_name == u'valid@home.org' )

        vassert( UserInput.query.filter_by(key=u'field_name').count() == 0 ) # The invalid input was removed
        vassert( PersistedException.query.count() == 0 ) # The exception was removed

        vassert( browser.location_path == u'/page2' )

        # Case: form validation passes (js)
        #  - no ValidationException
        #  - all input is translated to python and set as values on the model objects
        fixture.driver_browser.type(u'//input[@type="text"]', u'valid@home.org')
        fixture.driver_browser.wait_for_element_not_visible(fixture.error_xpath)
        fixture.driver_browser.click(u"//input[@value='click me']")
        vassert( model_object.field_name == u'valid@home.org' )

        vassert( fixture.driver_browser.current_url.path == u'/page2' )



    class QueryStringScenarios(FormFixture):
        def new_other_view(self):
            return self.wsgi_app.define_view(u'/page2', title=u'Page 2')

        def new_form(self):
            form = Form(self.view, u'some_form')
            event = Event(label=u'click me', action=Action(self.action))
            event.bind(u'an_event', None)
            form.define_event_handler(event, target=self.target)
            form.add_child(Button(form, event))
            return form

        @property
        def query_string_on_form_submit(self):
            form_action = self.driver_browser.get_attribute(u'//form', u'action')
            return Url(form_action).query

        @scenario
        def different_target(self):
            self.break_on_submit = False
            self.target_is_other_view = True
            self.initial_qs = u'a=b'
            self.final_qs = u''

        @scenario
        def same_target_by_intent(self):
            self.break_on_submit = False
            self.target_is_other_view = False
            self.target = self.view.as_factory()
            self.initial_qs = u'a=b'
            self.final_qs = u'a=b'

        @scenario
        def same_target_by_exception(self):
            self.break_on_submit = True
            self.target_is_other_view = True
            self.initial_qs = u'a=b'
            self.final_qs = u'a=b'


    @test(QueryStringScenarios)
    def propagation_of_querystring(self, fixture):
        """The query string of the original View visited is maintained through the GET/POST cycle.
           If, for whatever reason (exception or intent) the browser should stay on the same
           View after POSTing, the query string is maintained to that point.
           Else, the query string is cleared.
        """
        class ModelObject(object):
            def handle_event(self):
                if fixture.break_on_submit:
                    raise DomainException()
            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me', action=Action(self.handle_event))

        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name, other_view):
                super(MyForm, self).__init__(view, name)
                if fixture.target_is_other_view:
                    target = other_view
                else:
                    target = view.as_factory()
                self.define_event_handler(model_object.events.an_event, target=target)
                self.add_child(Button(self, model_object.events.an_event))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/', title=u'Home page')
                other_view = self.define_view(u'/page2', title=u'Page 2')
                home.set_slot(u'main', MyForm.factory(u'myform', other_view))

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        fixture.reahl_server.set_app(wsgi_app)
        fixture.driver_browser.open('/?%s' % fixture.initial_qs)

        vassert( fixture.query_string_on_form_submit == fixture.initial_qs )
        fixture.driver_browser.click(u"//input[@value='click me']")
        vassert( fixture.driver_browser.current_url.query == fixture.final_qs )


    @test(FormFixture)
    def event_names_are_canonicalised(self, fixture):
        """The name= attribute of a button is an url encoded string. There is more than one way
           to url encode the same string. The server ensures that different encodings of the same
           string are not mistaken for different names.
        """

        class ModelObject(object):
            def handle_event(self, some_argument):
                self.received_argument = some_argument

            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me',
                                        action=Action(self.handle_event, [u'some_argument']),
                                        some_argument=Field(default=u'default value'))

        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                self.define_event_handler(model_object.events.an_event)
                self.add_child(Button(self, model_object.events.an_event.with_arguments(some_argument=u'f~nnystuff')))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/', title=u'Home page')
                home.set_slot(u'main', MyForm.factory(u'myform'))

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # when the Action is executed, the correct arguments are passed
        browser.post(u'/__myform_method', {u'event.an_event?some_argument=f~nnystuff': ''})
        vassert( model_object.received_argument == u'f~nnystuff' )
        
    @test(FormFixture)
    def alternative_event_trigerring(self, fixture):
        """Events can also be triggered by submitting a Form via Ajax. In such cases the normal redirect-after-submit
           behaviour of the underlying EventChannel is not desirable. This behaviour can be switched off by submitting
           an extra argument along with the Form in order to request the alternative behaviour.
        """

        class ModelObject(object):
            def handle_event(self):
                self.handled_event = True

            @exposed
            def events(self, events):
                events.an_event = Event(label=u'click me',
                                        action=Action(self.handle_event))


        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name, other_view):
                super(MyForm, self).__init__(view, name)
                self.define_event_handler(model_object.events.an_event, target=other_view)
                self.add_child(Button(self, model_object.events.an_event))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(TwoColumnPage)
                home = self.define_view(u'/', title=u'Home page')
                other_view = self.define_view(u'/page2', title=u'Page 2')
                home.set_slot(u'main', MyForm.factory(u'myform', other_view))

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # when POSTing with _noredirect, the Action is executed, but the browser is not redirected to /page2 as usual
        browser.post(u'/__myform_method', {u'event.an_event?': u'', u'_noredirect': u''})
        browser.follow_response()  # Needed to make the test break should a HTTPTemporaryRedirect response be sent
        vassert( model_object.handled_event )
        vassert( browser.location_path != '/page2' )  
        vassert( browser.location_path == '/__myform_method' )  
        
        # the response is a json object reporting the success of the event and a new rendition of the form
        json_dict = json.loads(browser.raw_html)
        vassert( json_dict[u'success'] )

        browser.open('/')
        expected_html = browser.get_inner_html_for(u'//form[1]')
        vassert( json_dict[u'widget'].startswith(expected_html+u'<script') )


    @test(FormFixture)
    def remote_field_validation(self, fixture):
        """A Form contains a RemoteMethod that can be used to validate any of its fields via HTTP.
        """

        class ModelObject(object):
            @exposed
            def fields(self, fields):
                fields.a_field = EmailField()

        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                self.add_child(TextInput(self, model_object.fields.a_field))

        wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'))
        browser = Browser(wsgi_app)

        browser.open(u'/_myform_validate_method?a_field=invalid email address')
        vassert( browser.raw_html == u'"a_field should be a valid email address"' )

        browser.open(u'/_myform_validate_method?a_field=valid@email.org')
        vassert( browser.raw_html == u'true' )
        

    @test(FormFixture)
    def remote_field_formatting(self, fixture):
        """A Form contains a RemoteMethod that can be used to reformat any of its fields via HTTP.
        """

        class ModelObject(object):
            @exposed
            def fields(self, fields):
                fields.a_field = DateField()

        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                self.add_child(TextInput(self, model_object.fields.a_field))

        wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'))
        browser = Browser(wsgi_app)

        browser.open(u'/_myform_format_method?a_field=13 November 2012')
        vassert( browser.raw_html == u'13 Nov 2012' )

        browser.open(u'/_myform_format_method?a_field=invaliddate')
        vassert( browser.raw_html == u'' )











